############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Forrest Williams, Mar 2021                       #
############################################################


import datetime as dt
import os

import numpy as np

from mintpy.utils import readfile, utils1 as ut, writefile


#########################################################################
def add_cosicorr_metadata(fname, date12_dict, meta):
    '''Read/extract attribute data from cosicorr metadata file and add to metadata dictionary
    Parameters: fname       - str, offset or SNR file name
                date12_dict - dict, key for file name, value for date12 pairs
                meta        - dict, metadata dictionary
    Returns:    meta        - dict, metadata dictionary
    '''

    # add general attributes
    meta['PROCESSOR'] = 'cosicorr'
    meta['P_BASELINE_TOP_HDR'] = 0.0 #placeholder
    meta['P_BASELINE_BOTTOM_HDR'] = 0.0 #placeholder
    meta['RANGE_PIXEL_SIZE'] = np.abs(meta['X_STEP'])
    meta['AZIMUTH_PIXEL_SIZE'] = np.abs(meta['Y_STEP'])
    meta['RLOOKS'] = 1
    meta['ALOOKS'] = 1

    # Time attributes
    date1_str, date2_str = date12_dict[os.path.basename(fname)].split('-')
    meta['DATE12'] = f'{date1_str}-{date2_str}'
    date1 = dt.datetime.strptime(date1_str, '%Y%m%d')
    date2 = dt.datetime.strptime(date2_str, '%Y%m%d')
    date_avg = date1 + (date2 - date1) / 2
    date_avg_seconds = (date_avg - date_avg.replace(hour=0, minute=0, second=0, microsecond=0)).total_seconds()
    meta['CENTER_LINE_UTC'] = date_avg_seconds

    # add LAT/LON_REF1/2/3/4
    N = float(meta['Y_FIRST'])
    W = float(meta['X_FIRST'])
    S = N + float(meta['Y_STEP']) * int(meta['LENGTH'])
    E = W + float(meta['X_STEP']) * int(meta['WIDTH'])

    meta['LAT_REF1'] = str(S)
    meta['LAT_REF2'] = str(S)
    meta['LAT_REF3'] = str(N)
    meta['LAT_REF4'] = str(N)
    meta['LON_REF1'] = str(W)
    meta['LON_REF2'] = str(E)
    meta['LON_REF3'] = str(W)
    meta['LON_REF4'] = str(E)

    return(meta)


def prep_cosicorr(inps):
    """Prepare the COSI-Corr metadata."""

    # open and read hyp3 metadata
    date12_dict = {}
    with open(inps.meta_file) as f:
        for line in f:
            name, date1, date2 = line.strip().split(' ')
            date12_dict[name] = f'{date1}-{date2}'

    # loop over each filename
    inps.file = ut.get_file_list(inps.file, abspath=True)
    for fname in inps.file:
        # extra metadata
        meta = readfile.read_gdal_vrt(fname)
        meta = add_cosicorr_metadata(fname, date12_dict, meta)

        # write metadata into RSC file
        rsc_file = fname+'.rsc'
        writefile.write_roipac_rsc(meta, out_file=rsc_file)

    return



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Antonio Valentino, Zhang Yunjun, Aug 2022        #
############################################################


import sys

from mintpy.utils.arg_utils import create_argument_parser

##################################################################################################
EXAMPLE = """example:
  multilook.py  velocity.h5  -r 15 -a 15
  multilook.py  srtm30m.dem  -r 10 -a 10  -o srtm30m_300m.dem

  # Ignore / skip marginal pixels
  multilook.py ../../geom_reference/hgt.rdr.full -r 300 -a 100 --margin 58 58 58 58 -o hgt.rdr
"""


def create_parser(subparsers=None):
    synopsis = 'Multilook the input file'
    epilog = EXAMPLE
    name = __name__.split('.')[-1]
    parser = create_argument_parser(
        name, synopsis=synopsis, description=synopsis, epilog=epilog, subparsers=subparsers)

    parser.add_argument('file', nargs='+', help='File(s) to multilook')
    parser.add_argument('-r','--range','-x', dest='lks_x', type=int, default=1,
                        help='number of multilooking in range  /x direction (default: %(default)s).')
    parser.add_argument('-a','--azimuth','-y', dest='lks_y', type=int, default=1,
                        help='number of multilooking in azimuth/y direction (default: %(default)s).')
    parser.add_argument('-o', '--outfile',
                        help='Output file name. Disabled when more than 1 input files')
    parser.add_argument('-m','--method', dest='method', type=str, default='mean', choices=['mean', 'median', 'nearest'],
                        help='downsampling method (default: %(default)s) \n'
                             'e.g. nearest for geometry, average for observations')
    parser.add_argument('--margin', dest='margin', type=int, nargs=4, metavar=('TOP','BOTTOM','LEFT','RIGHT'),
                        default=[0,0,0,0], help='number of pixels on the margin to skip, (default: %(default)s).')
    return parser


def cmd_line_parse(iargs=None):
    # parse
    parser = create_parser()
    inps = parser.parse_args(args=iargs)

    # import
    from mintpy.utils import utils1 as ut

    # check: input file list
    inps.file = ut.get_file_list(inps.file)

    # check: -x/y options (num of multilooks)
    if inps.lks_x == 1 and inps.lks_y == 1:
        raise SystemExit('ERROR: no multilooking specified: lks_x/y=1!')

    # check: -o / --outfile (output file name)
    if len(inps.file) > 1 and inps.outfile:
        inps.outfile = None
        print('more than one file is input, disable custom output filename.')

    return inps


##################################################################################################
def main(iargs=None):
    # parse
    inps = cmd_line_parse(iargs)

    # import
    from mintpy.multilook import multilook_file

    # run
    for infile in inps.file:
        multilook_file(
            infile,
            lks_y=inps.lks_y,
            lks_x=inps.lks_x,
            outfile=inps.outfile,
            method=inps.method,
            margin=inps.margin)
    print('Done.')


###################################################################################################
if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3
#############################################################
# Program is part of MintPy                                 #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi          #
# Author: Joshua Zahner, David Grossman, Zhang Yunjun, 2020 #
#############################################################
# Recommend import:
#     from mintpy.objects import cluster


import glob
import os
import shutil
import time

# supported / tested clusters
CLUSTER_LIST = ['lsf', 'pbs', 'slurm', 'local']
NUM_THREADS_ENV_LIST = [
    'OMP_NUM_THREADS',         # openmp
    'OPENBLAS_NUM_THREADS',    # openblas
    'MKL_NUM_THREADS',         # mkl
    'NUMEXPR_NUM_THREADS',     # numexpr
    'VECLIB_MAXIMUM_THREADS',  # accelerate
]


############################## Utilities functions #########################################

def split_box2sub_boxes(box, num_split, dimension='x', print_msg=False):
    """Divide the input box into `num_split` different sub_boxes.

    :param box: [x0, y0, x1, y1]: list[int] of size 4
    :param num_split: int, the number of sub_boxes to split a box into
    :param dimension: str = 'y' or 'x', the dimension along which to split the boxes
    :return: sub_boxes: list(list(4 int)), the splited sub boxes
    """
    import numpy as np

    dimension = dimension.lower()
    if num_split <= 1:
        return [box]

    # basic info
    x0, y0, x1, y1 = box
    length, width = y1 - y0, x1 - x0

    # calc step
    if dimension == 'y':
        dim_size = length
    else:
        dim_size = width
    step = int(np.ceil(dim_size / num_split))
    step = max(step, 10)                       # constain the min step size
    num_split = int(np.ceil(dim_size / step))  # trim the final number of boxes

    # get list of boxes
    sub_boxes = []
    for i in range(num_split):
        if dimension == 'y':
            r0 = y0 + step * i
            r1 = y0 + step * (i + 1)
            r1 = min(r1, y1)
            sub_boxes.append([x0, r0, x1, r1])

        else:
            c0 = x0 + step * i
            c1 = x0 + step * (i + 1)
            c1 = min(c1, x1)
            sub_boxes.append([c0, y0, c1, y1])

    if print_msg:
        print(f'split along {dimension} dimension ({dim_size:d}) into {num_split:d} boxes')
        print(f'    with each box up to {step:d} in {dimension} dimension')

    return sub_boxes


def set_num_threads(num_threads=None, print_msg=True):
    """limit/set the number of threads for all environmental variables to the given value
    and save/return the original value for backup purpose.
    Link: https://stackoverflow.com/questions/30791550

    Parameters: num_threads      - str, number of threads
                                   Set to None to return without changing env variables
    Returns:    num_threads_dict - dict, dictionary of the original number of threads
    """

    # grab the original number of threads
    if print_msg:
        print(f'save the original settings of {NUM_THREADS_ENV_LIST}')
    num_threads_dict = {}
    for key in NUM_THREADS_ENV_LIST:
        num_threads_dict[key] = os.environ.get(key, None)

    # change the env variables
    if num_threads:
        num_threads = str(num_threads)
        for key in NUM_THREADS_ENV_LIST:
            os.environ[key] = num_threads
            if print_msg:
                print(f'set {key} = {num_threads}')

    return num_threads_dict


def roll_back_num_threads(num_threads_dict, print_msg=True):
    """Set back the number of threads for all environmental variables."""
    if print_msg:
        print(f'roll back to the original settings of {NUM_THREADS_ENV_LIST}')
    for key, value in num_threads_dict.items():
        if key in os.environ.keys():
            if value is None:
                os.environ.pop(key)
                if print_msg:
                    print(f'remove env variable {key}')
            else:
                os.environ[key] = value
                if print_msg:
                    print(f'set {key} = {value}')
    return



############################## Beginning of DaskCluster class ##############################

class DaskCluster:
    """
    Generic dask cluster wrapper for parallel processing in blocks.

    This object takes in a computing function for one block in space.
    For the computing function:
        1. the output is always several matrices (or None) and one box.
        2. the number of matrices may vary for different applications/functions.
        3. all matrices will be in 2D in size of (len, wid) or 3D in size of (n, len, wid),
           thus, the last two dimension (in space) will be the same.
    This charateristics allows the automatic result collection without prior knowledge
        of the computing funciton, thus being a generic wrapper.

    Check ifgram_inversion.py as an example.

    """

    def __init__(self, cluster_type, num_worker, config_name=None, **kwargs):
        """Initiate object
        :param cluster_type: str, cluster to use (local, slurm, lsf, pbs)
        :param num_worker: str, number of workers to use
        :param config_name: str, the name of configuratino section
        :other param **kwargs: dask configuration parameters
                 e.g. config_name: str, the user specified config name to use
        """

        self.cluster_type = cluster_type.lower()
        self.num_worker = num_worker
        self.config_name = config_name
        self.cluster_kwargs = kwargs

        ## format input arguments
        # num_worker
        self.num_worker = self.format_num_worker(self.cluster_type, self.num_worker)

        # config_name
        self.format_config_name()
        self.cluster_kwargs['config_name'] = self.config_name

        ## printout message
        print(f"input Dask cluster type: {self.cluster_type}")
        if self.config_name is not None:
            print(f"input Dask config name: {self.config_name}")

        ## intitial value
        self.cluster = None
        self.client = None


    def open(self):
        """Initiate the cluster"""

        # initiate the cluster object
        # Look at the ~/.config/dask/mintpy.yaml file for changing the Dask configuration defaults
        print('initiate Dask cluster')
        if self.cluster_type == 'local':
            from dask.distributed import LocalCluster

            # initiate cluster object
            self.cluster = LocalCluster()

        else:
            # for non-local cluster, import related dask module only when it's needed
            # because job_queue is not available on macports, which make sense
            import dask_jobqueue

            # initiate cluster object
            if self.cluster_type == 'lsf':
                self.cluster = dask_jobqueue.LSFCluster(**self.cluster_kwargs)

            elif self.cluster_type == 'pbs':
                self.cluster = dask_jobqueue.PBSCluster(**self.cluster_kwargs)

            elif self.cluster_type == 'slurm':
                self.cluster = dask_jobqueue.SLURMCluster(**self.cluster_kwargs)

            else:
                msg = f'un-recognized input cluster: {self.cluster_type}'
                msg += f'\nsupported clusters: {CLUSTER_LIST}'
                raise ValueError(msg)

            # show dask cluster job script for reference
            print("\n", self.cluster.job_script())
            # for debug
            debug_mode = False
            if debug_mode:
                with open('dask_command_run_from_python.txt', 'w') as f:
                    f.write(self.cluster.job_script() + '\n')


    def run(self, func, func_data, results):
        """Wrapper function encapsulating submit_workers and compile_workers.

        For a generic result collection without prior knowledge of the computing function,
        we assume that the output of "func" is: several 2D or 3D matrices + a box

        :param func: function, a python function to run in parallel
        :param func_data: dict, a dictionary of the argument to pass to the function
        :param results: list[numpy.ndarray], arrays of the appropriate structure representing
               the final output of processed box (need to be in the same order as the function passed in
               submit_workers returns in)
        :return: results: tuple(numpy.ndarray), the processed results of the box
        """
        from dask.distributed import Client

        # split the primary box into sub boxes for workers AND
        # update the number of workers based on split result
        box = func_data["box"]
        sub_boxes = split_box2sub_boxes(box, num_split=self.num_worker, dimension='x', print_msg=False)
        self.num_worker = len(sub_boxes)
        print(f'split patch into {self.num_worker} sub boxes in x direction for workers to process')

        # start a bunch of workers from the cluster
        print(f'scale Dask cluster to {self.num_worker} workers')
        self.cluster.scale(self.num_worker)

        print('initiate Dask client')
        self.client = Client(self.cluster)
        self.client.get_versions(check=True)

        # submit job for each worker
        futures, submission_time = self.submit_job(func, func_data, sub_boxes)

        # assemble results from all workers
        results = self.collect_result(futures, results, box, submission_time)

        return results


    def submit_job(self, func, func_data, sub_boxes):
        """Submit dask workers to the networking client that run the specified function (func)
        on the specified data (func_data). Each dask worker is in charge of a small subbox of the main box.

        :param func: function, a python function to run in parallel
        :param func_data: dict, a dictionary of the argument to pass to the function
        :param sub_boxes: list(np.ndarray), list of boxes to be computed in parallel

        :return futures: list(dask.Future), list of futures representing future dask worker calculations
        :return submission_time: time, the time of submission of the dask workers (used to determine worker
                runtimes as a performance diagnostic)
        """

        submission_time = time.time()
        futures = []
        for i, sub_box in enumerate(sub_boxes):
            print(f'submit a job to the worker for sub box {i}: {sub_box}')
            func_data['box'] = sub_box

            # David: I haven't played with fussing with `retries`, however sometimes a future fails
            # on a worker for an unknown reason. retrying will save the whole process from failing.
            # TODO:  I don't know what to do if a future fails > 3 times. I don't think an error is
            # thrown in that case, therefore I don't know how to recognize when this happens.
            future = self.client.submit(func, **func_data, retries=3)
            futures.append(future)

        return futures, submission_time


    def collect_result(self, futures, results, box, submission_time):
        """Compile results from completed workers and recompiles their sub outputs into the output
        for the complete box being worked on.
        :param futures: list(dask.Future), list of futures representing future dask worker calculations
        :param results: list[numpy.ndarray], arrays of the appropriate structure representing
               the final output of processed box (need to be in the same order as the function passed in
               submit_workers returns in)
        :param box: numpy.ndarray, the initial complete box being processed
        :param submission_time: time, the time of submission of the dask workers (used to determine worker
               runtimes as a performance diagnostic)
        :return: results: tuple(numpy.ndarray), the processed results of the box
        """
        from dask.distributed import as_completed

        num_future = 0
        for future, sub_results in as_completed(futures, with_results=True):

            # message
            num_future += 1
            sub_t = time.time() - submission_time
            print(f"\nFUTURE #{num_future} complete. Time used: {sub_t:.0f} seconds")

            # catch result - sub_box
            # and convert the abosulte sub_box into local col/row start/end relative to the primary box
            # to assemble the result from each worker
            sub_box = sub_results[-1]
            x0, y0, x1, y1 = sub_box
            x0 -= box[0]
            x1 -= box[0]
            y0 -= box[1]
            y1 -= box[1]

            # catch result - matrices
            # and loop across all of the returned data to rebuild complete box
            for i, sub_result in enumerate(sub_results[:-1]):
                if sub_result is not None:
                    num_dim = sub_result.ndim
                    if num_dim == 4:
                        results[i][:, :,
                                   y0:y1, x0:x1] = sub_result
                    elif num_dim == 3:
                        results[i][:,
                                   y0:y1, x0:x1] = sub_result
                    elif num_dim == 2:
                        results[i][y0:y1, x0:x1] = sub_result
                    else:
                        msg = f"worker result has unexpected dimension: {num_dim}"
                        msg += '\nit should be either 2 or 3 or 4!'
                        raise Exception(msg)

        return results


    def close(self):
        """Close connections to dask client and cluster and moves dask output/error files. """

        # close client before cluster -> less likely to have the CancelledError
        # https://github.com/dask/distributed/issues/2273
        self.client.close()
        print('close dask client')

        self.cluster.close()
        print('close dask cluster')

        # move *.o/.e files produced by dask in stdout/stderr
        self.move_dask_stdout_stderr_files()


    ##### Utilities functions

    @staticmethod
    def format_num_worker(cluster_type, num_worker):
        """Format dask num_worker.
        :param cluster_type: str
        :param num_worker: str, number of workers to use
        :return: num_worker: int, number of workers to use
        """

        if cluster_type == 'local':
            num_core = os.cpu_count()

            # all / percentage --> num_core
            msg = f'numWorker = {num_worker}'
            if num_worker == 'all':
                ## divide by the number of threads per core [for Linux only]
                #import subprocess
                #from mintpy.utils import utils0 as ut0
                #if ut0.which('lscpu') is not None:
                #    # get the number of threads per core
                #    # link: https://stackoverflow.com/questions/62652951
                #    ps = subprocess.run(['lscpu'], capture_output=True, text=True).stdout.split('\n')
                #    ns = [p.split(':')[1].strip() for p in ps if p.startswith('Thread(s) per core:')]
                #    if len(ns) > 0:
                #        num_thread = int(ns[0])
                #        num_core = int(num_core / num_thread)

                # set num_worker to the number of cores
                num_worker = str(num_core)
                print(f'translate {msg} to {num_worker}')

            elif num_worker.endswith('%'):
                num_worker = int(num_core * float(num_worker[:-1]) / 100)
                print(f'translate {msg} to {num_worker}')
                if num_worker < 1 or num_worker >= num_core:
                    raise ValueError('Invalid numWorker percentage!')

            # str --> int
            num_worker = int(num_worker)

            # if num_worker > num_core,
            # then we assume that the user is not aware of the available resources
            # and use max(num_core/2, 1) instead to be conservative.
            if num_worker > num_core:
                print(f'\nWARNING: input number of worker: {num_worker} > available cores: {num_core}')
                num_worker = max(int(num_core / 2), 1)
                print(f'change number of worker to {num_worker} and continue\n')

        else:
            if num_worker == 'all':
                msg = f'numWorker = all is NOT supported for cluster type: {cluster_type}'
                raise ValueError(msg)
            num_worker = int(num_worker)

        return num_worker


    def format_config_name(self):
        """Format dask config_name property based on presence or absence of user specified config name.

        :return: config_name: str, the config_name formatted as follows:
                 - the user specified config name if its exists in $DASK_CONFIG/dask.yaml
                 - the default cluster_type config in $DASK_CONFIG/dask.yaml
        """
        import dask

        # config_name is not needed for local cluster
        if self.cluster_type == 'local':
            self.config_name = None
            return self.config_name

        # translate config_name = None to config_name = cluster_type
        if self.config_name is None:
            print('input config name is None, thus use the default (same as cluster type)')
            self.config_name = self.cluster_type

        # if config_name not found, use cluster_type as defined in minpy.dask
        config_names = list(dask.config.get('jobqueue').keys())
        if self.config_name not in config_names:
            config_location = dask.config.get('config')
            msg = f'Dask configuration "{self.config_name}" was not found in {config_location}'
            msg += f'\nFalling back to default config name: "{self.cluster_type}"'
            print(msg)
            self.config_name = self.cluster_type

        return self.config_name


    def move_dask_stdout_stderr_files(self):
        """Move *o and *e files produced by dask into stdout and sderr directory"""

        stdout_files = glob.glob('*.o')
        stderr_files = glob.glob('*.e')
        job_files = glob.glob('dask_command_run_from_python.txt*')

        if len(stdout_files + stderr_files + job_files) == 0:
            return

        stdout_folder = 'stdout_dask'
        stderr_folder = 'stderr_dask'
        for std_dir in [stdout_folder, stderr_folder]:
            if os.path.isdir(std_dir):
                shutil.rmtree(std_dir)
            os.mkdir(std_dir)

        for item in stdout_files + job_files:
            shutil.move(item, stdout_folder)

        for item in stderr_files:
            shutil.move(item, stderr_folder)

############################## End of DaskCluster class ####################################



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Heresh Fattahi, 2013                             #
############################################################


import os
import sys

import h5py
import matplotlib.pyplot as plt
import numpy as np
from scipy import io as sio, ndimage

from mintpy.utils import readfile


############################################################
def dms2d(Coord):
    d, m, s = Coord.split(' ')
    d = float(d)
    m = float(m)
    s = float(s)
    d = d+m/60.+s/3600.
    return d


############################################################
def gps_to_LOS(Ve, Vn, theta, heading):
    unitVec = [np.cos(heading)*np.sin(theta), -np.sin(theta)
               * np.sin(heading), np.cos(theta)]
    gpsLOS = unitVec[0]*Ve+unitVec[1]*Vn
    return gpsLOS


############################################################
def check_st_in_box(x,y,x0,y0,x1,y1,X0,Y0,X1,Y1):

    m1=float(y1-y0)/float(x1-x0)
    c1=float(y0-m1*x0)

    m2=float(Y1-Y0)/float(X1-X0)
    c2=float(Y0-m2*X0)

    m3=float(y0-Y0)/float(x0-X0)
    c3=float(Y0-m3*X0)

    m4=float(y1-Y1)/float(x1-X1)
    c4=float(Y1-m4*X1)


    yy1=m1*x+c1
    yy2=m2*x+c2

    yy=[yy1,yy2]
    yy.sort()

    xx3=(y-c3)/m3
    xx4=(y-c4)/m4
    xx=[xx3,xx4]
    xx.sort()
    if y >= yy[0] and y <= yy[1] and x >= xx[0] and x <= xx[1]:
        Check_result = 'True'
    else:
        Check_result='False'


    return Check_result


############################################################
def check_st_in_box2(x, y, x0, y0, x1, y1, X0, Y0, X1, Y1):

    m1, c1 = line(x0, y0, x1, y1)
    m2, c2 = line(X0, Y0, X1, Y1)
    m3, c3 = line(x0, y0, X0, Y0)
    m4, c4 = line(x1, y1, X1, Y1)

    D1 = np.sqrt((x1-x0)**2+(y1-y0)**2)
    D2 = np.sqrt((X0-x0)**2+(Y0-y0)**2)

    d1 = dist_point_from_line(m1, c1, x, y, 1, 1)
    d2 = dist_point_from_line(m2, c2, x, y, 1, 1)

    d3 = dist_point_from_line(m3, c3, x, y, 1, 1)
    d4 = dist_point_from_line(m4, c4, x, y, 1, 1)

    if np.round(d1+d2) == np.round(D2) and np.round(d3+d4) == np.round(D1):
        Check_result = 'True'
    else:
        Check_result = 'False'

    return Check_result


############################################################
def line(x0, y0, x1, y1):
    m = float(y1-y0)/float(x1-x0)
    c = float(y0-m*x0)
    return m, c


############################################################
def dist_point_from_line(m, c, x, y, dx, dy):
    # finds the distance of a point at x ,y xoordinate
    # from a line with Y =  mX +c
    d = np.sqrt((((x+m*y-m*c)/(m**2+1)-x)*dx)**2 +
                ((m*(x+m*y-m*c)/(m**2+1)+c-y)*dy)**2)

    #  a=m;b=-1;
    # d=np.abs(a*x+b*y+c)/np.sqrt(a**2+b**2)
    return d


############################################################
def get_intersect(m, c, x, y):

    xp = (x+m*y-m*c)/(m**2+1)
    yp = m*(x+m*y-m*c)/(m**2+1)+c
    return xp, yp


############################################################
def readGPSfile(gpsFile, gps_source):
    if gps_source in ['cmm4', 'CMM4']:

        gpsData = np.loadtxt(gpsFile, usecols=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(0, 1))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []
        for i in range(gpsData.shape[0]):
            if 'GPS' in Stations[i]:
                Lat.append(gpsData[i, 0])
                Lon.append(gpsData[i, 1])
                Ve.append(gpsData[i, 2])
                Se.append(gpsData[i, 3])
                Vn.append(gpsData[i, 4])
                Sn.append(gpsData[i, 5])
                St.append(Stations[i])

    elif gps_source == 'mintpy':

        gpsData = np.loadtxt(gpsFile, usecols=(1, 2, 3, 4, 5, 6))  # ,7,8,9))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(0, 1))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []

        for i in range(gpsData.shape[0]):
            Lon.append(gpsData[i, 0])
            Lat.append(gpsData[i, 1])
            Ve.append(gpsData[i, 2])
            Vn.append(gpsData[i, 3])
            Se.append(gpsData[i, 4])
            Sn.append(gpsData[i, 5])
            St.append(Stations[i])

    elif gps_source in ['usgs', 'USGS']:

        gpsData_Hz = np.loadtxt(gpsFile, usecols=(0, 1, 2, 3, 4, 5, 6))
        gpsData_up = np.loadtxt(gpsFile, usecols=(8, 9))
        gpsData = np.hstack((gpsData_Hz, gpsData_up))
        Stations = np.loadtxt(gpsFile, dtype=str, usecols=(7, 8))[:, 0]

        St = []
        Lon = []
        Lat = []
        Ve = []
        Vn = []
        Se = []
        Sn = []

        for i in range(gpsData.shape[0]):
            Lat.append(gpsData[i, 0])
            Lon.append(gpsData[i, 1])
            Vn.append(gpsData[i, 2])
            Ve.append(gpsData[i, 3])
            Sn.append(gpsData[i, 4])
            Se.append(gpsData[i, 5])
            St.append(Stations[i])

    return list(St), Lat, Lon, Ve, Se, Vn, Sn


############################################################
def redGPSfile(gpsFile):
    gpsData_Hz = np.loadtxt(gpsFile, usecols=(0, 1, 2, 3, 4, 5, 6))
    gpsData_up = np.loadtxt(gpsFile, usecols=(8, 9))
    gpsData = np.hstack((gpsData_Hz, gpsData_up))
    Stations = np.loadtxt(gpsFile, dtype=str, usecols=(7, 8))[:, 0]
    return list(Stations), gpsData


############################################################
def redGPSfile_cmm4(gpsFile):
    gpsData = np.loadtxt(gpsFile, usecols=(1, 2, 3, 4, 5, 6, 7, 8, 9))
    Stations = np.loadtxt(gpsFile, dtype=str, usecols=(0, 1))[:, 0]
    return list(Stations), gpsData


############################################################
def nearest(x, tbase, xstep):
    ## """ find nearest neighbour """
    dist = np.sqrt((tbase - x)**2)
    if min(dist) <= np.abs(xstep):
        indx = dist == min(dist)
    else:
        indx = []
    return indx


############################################################
def find_row_column(Lon, Lat, lon, lat, lon_step, lat_step):
    # finding row and column numbers of the GPS point

    idx = nearest(Lon, lon, lon_step)
    idy = nearest(Lat, lat, lat_step)
    if idx != [] and idy != []:
        IDX = np.where(idx == True)[0][0]
        IDY = np.where(idy == True)[0][0]
    else:
        IDX = np.nan
        IDY = np.nan
    return IDY, IDX


################################################
def get_lat_lon(atr):
    Width = float(atr['WIDTH'])
    Length = float(atr['LENGTH'])
    ullon = float(atr['X_FIRST'])
    ullat = float(atr['Y_FIRST'])
    lon_step = float(atr['X_STEP'])
    lat_step = float(atr['Y_STEP'])
    lon_unit = atr['Y_UNIT']
    lat_unit = atr['X_UNIT']

    lllat = ullat + (Length-1)*lat_step
    urlon = ullon + (Width - 1)*lon_step
    lat = np.linspace(ullat, lllat, Length)
    lon = np.linspace(ullon, urlon, Width)

    lon_all = np.tile(lon, (Length, 1))
    lat_all = np.tile(lat, (Width, 1)).T

    return lat, lon, lat_step, lon_step, lat_all, lon_all


############################################################
def nanmean(data, **args):
    return np.ma.filled(np.ma.masked_array(data, np.isnan(data)).mean(**args), fill_value=np.nan)


def nanstd(data, **args):
    return np.ma.filled(np.ma.masked_array(data, np.isnan(data)).std(**args), fill_value=np.nan)


############################################################
def get_transect(z, x0, y0, x1, y1, interpolation='nearest'):
    # Extract 2D matrxi z value along the line [x0,y0;x1,y1]
    # Ref link: http://stackoverflow.com/questions/7878398/how-to-e
    # xtract-an-arbitrary-line-of-values-from-a-numpy-array
    ##
    # Option: interpolation : sampling/interpolation method, including:
    # 'nearest'  - nearest neighbour, by default
    # 'cubic'    - cubic interpolation
    # 'bilinear' - bilinear interpolation
    ##

    # Extract the line
    length = int(np.hypot(x1-x0, y1-y0))
    x, y = np.linspace(x0, x1, length), np.linspace(y0, y1, length)

    ## Extract the value along the line
    if interpolation.lower() == 'cubic':
        zi = ndimage.map_coordinates(z, np.vstack((x, y)))
    elif interpolation.lower() == 'bilinear':
        zi = ndimage.map_coordinates(z, np.vstack((x, y)), order=2)
    else:
        zi = z[np.rint(y).astype(np.int), np.rint(x).astype(np.int)]     # nearest neighbour

    return zi


############################################################
USAGE = """
*****************************************************************************************

   Generating a transect [or multiple transects] of the velocity field.
   If GPS velocities are provided, it will be compared with InSAR.

   Usage:

       transect.py -f velocity.h5 -s 'y1,x1' -e 'y2,x2  -n number_of_transects  -d distace_between_profiles(pixel)
                   -g gps velocity file -r reference station -L List of stations

       -s : strat point of the profile
       -e : end   point of the profile
       -F : Fault coordinates (lat_first, lon_first, lat_end, lon_end)
       -n : number of transections                [default: 1]
       -d : distance between profiles in pixels   [default: 1]
       -p : flip profile left - right (yes or no) [default: no]
       -u : flip profile up   - down              [default: no]
       -S : source of GPS velocities (usgs,cmm4,mintpy)
       -G : gps stations to compare with InSAR  (all,insar,profile)
            "all"     : all gps stations is projected to the profile
            "insar"   : same as all but limited to the area covered by insar
            "profile" : only those gps stations which are in the profile area]

       -l : lower  bound to display
       -h : higher bound to display
       -I : display InSAR velocity [on] or off
       -A : display Average InSAR velocity [on] or off
       -U : display Standard deviation of the InSAR velocity [on] or off
       -E : Export the generated transect to a matlab file [off] or on

   Example:

       transect.py -f geo_velocity.h5
       transect.py -f geo_velocity.h5 -s '5290,5579' -e '12177,482'
       transect.py -f geo_velocity_New_masked_masked.h5 -s '3967,7019' -e '12605,1261' -n 150 -d 10
                   -g usgs_velocities_NAfixed.txt -S usgs -r P625
       transect.py -f geo_velocity_New_masked_masked.h5 -g usgs_velocities_NAfixed.txt
                   -r P625 -F '33 30 24,-115 20 15,33 35 21,-116 10 16' -s '12644,1183' -e '6512,6227'
                   -n 100 -d 10 -p no -S usgs  -G insar
       transect.py -f geo_velocity_demCor_tropCor_masked_masked.h5 -F '33 30 24,-115 20 15,33 35 21,-116 10 16'
                   -n 100 -d 10 -p yes -g ../GPS_velocities_PBO_and_unavco.cmm4 -S mintpy -r IMPS -n 230 -s '11904,93' -e '5389,5662' -E on

*****************************************************************************************
"""


def Usage():
    print(USAGE)


#####################################################################
def main(argv):
    dp = 1.0
    ntrans = 1
    save_to_mat = 'off'
    flip_profile = 'no'
    which_gps = 'all'
    flip_updown = 'yes'
    incidence_file = 'incidence_file'
    display_InSAR = 'on'
    display_Average = 'on'
    disp_std = 'on'

    ##### Input Args
    try:
        opts, args = getopt.getopt(argv,"f:s:e:n:d:g:l:h:r:L:F:p:u:G:S:i:I:A:U:E:")
    except getopt.GetoptError:
        Usage()
        sys.exit(1)

    for opt,arg in opts:
        if   opt == '-f':   velocityFile = arg
        elif opt == '-s':   y0,x0  = (int(i) for i in arg.split(','))
        elif opt == '-e':   y1,x1  = (int(i) for i in arg.split(','))
        elif opt == '-n':   ntrans = int(arg)
        elif opt == '-d':   dp     = float(arg)
        elif opt == '-g':   gpsFile=arg
        elif opt == '-r':   refStation=arg
        elif opt == '-i':   incidence_file=arg
        elif opt == '-L':   stationsList = arg.split(',')
        elif opt == '-F':   FaultCoords  = arg.split(',')
        elif opt == '-p':   flip_profile = arg
        elif opt == '-u':   flip_updown  = arg; print(flip_updown)
        elif opt == '-G':   which_gps =arg
        elif opt == '-S':   gps_source=arg
        elif opt == '-h':   hbound=float(arg)
        elif opt == '-l':   lbound=float(arg)
        elif opt == '-I':   display_InSAR   = arg
        elif opt == '-A':   display_Average = arg
        elif opt == '-U':   disp_std        = arg
        elif opt == '-E':   save_to_mat     = arg

    ##### Input File Info
    try:
        atr = readfile.read_attribute(velocityFile)
    except:
        Usage()
        sys.exit(1)
    k = atr['FILE_TYPE']
    print('input file is '+k)

    h5file = h5py.File(velocityFile, 'r')
    z = h5file[k].get(k)[:]

    width = int(atr['WIDTH'])
    length = int(atr['LENGTH'])
    try:
        lat, lon, lat_step, lon_step, lat_all, lon_all = get_lat_lon(atr)
    except:
        print('radar coordinate')

    # Fault Coordinates
    try:
        Lat0 = dms2d(FaultCoords[0])
        Lon0 = dms2d(FaultCoords[1])
        Lat1 = dms2d(FaultCoords[2])
        Lon1 = dms2d(FaultCoords[3])
        Length, Width = np.shape(z)
        Yf0, Xf0 = find_row_column(Lon0, Lat0, lon, lat, lon_step, lat_step)
        Yf1, Xf1 = find_row_column(Lon1, Lat1, lon, lat, lon_step, lat_step)

        print('*********************************************')
        print(' Fault Coordinates:')
        print('   --------------------------  ')
        print('    Lat          Lon')
        print(str(Lat0) + ' , ' + str(Lon0))
        print(str(Lat1) + ' , ' + str(Lon1))
        print('   --------------------------  ')
        print('    row          column')
        print(str(Yf0) + ' , ' + str(Xf0))
        print(str(Yf1) + ' , ' + str(Xf1))
        print('*********************************************')
        # mf=float(Yf1-Yf0)/float((Xf1-Xf0))  # slope of the fault line
        # cf=float(Yf0-mf*Xf0)   # intercept of the fault line
        # df0=dist_point_from_line(mf,cf,x0,y0,1,1)   #distance of the profile start point from the Fault line
        # df1=dist_point_from_line(mf,cf,x1,y1,1,1)  #distance of the profile end point from the Fault line

        # mp=-1./mf  # slope of profile which is perpendicualr to the fault line
        # correcting the end point of the profile to be on a line perpendicular to the Fault
        # x1=int((df0+df1)/np.sqrt(1+mp**2)+x0)
        # y1=int(mp*(x1-x0)+y0)
    except:
        print('*********************************************')
        print('No information about the Fault coordinates!')
        print('*********************************************')

    #############################################################################
    try:
        x0
        y0
        x1
        y1
    except:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.imshow(z)
        try:
            ax.plot([Xf0, Xf1], [Yf0, Yf1], 'k-')
        except:
            print('Fault line is not specified')

        xc = []
        yc = []
        print('please click on start and end point of the desired profile')
        def onclick(event):
            if event.button == 1:
                print('click')
                xc.append(int(event.xdata))
                yc.append(int(event.ydata))
        cid = fig.canvas.mpl_connect('button_press_event', onclick)
        plt.show()
        x0 = xc[0]
        x1 = xc[1]
        y0 = yc[0]
        y1 = yc[1]
    ##############################################################################
    try:
        mf = float(Yf1-Yf0)/float(Xf1-Xf0)  # slope of the fault line
        cf = float(Yf0-mf*Xf0)   # intercept of the fault line
        # distance of the profile start point from the Fault line
        df0 = dist_point_from_line(mf, cf, x0, y0, 1, 1)
        # distance of the profile end point from the Fault line
        df1 = dist_point_from_line(mf, cf, x1, y1, 1, 1)

        mp = -1./mf  # slope of profile which is perpendicualr to the fault line
        # correcting the end point of the profile to be on a line perpendicular to the Fault
        x1 = int((df0+df1)/np.sqrt(1+mp**2)+x0)
        y1 = int(mp*(x1-x0)+y0)
    except:
        Info_aboutFault = 'No'

    ##############################################################################
    print('******************************************************')
    print('First profile coordinates:')
    print('Start point:  y = '+str(y0)+', x = '+str(x0))
    print('End   point:  y = '+str(y1)+', x = '+str(x1))
    print('******************************************************')
    length = int(np.hypot(x1-x0, y1-y0))
    x, y = np.linspace(x0, x1, length), np.linspace(y0, y1, length)
    zi = z[y.astype(np.int), x.astype(np.int)]
    try:
        lat_transect = lat_all[y.astype(np.int), x.astype(np.int)]
        lon_transect = lon_all[y.astype(np.int), x.astype(np.int)]
    except:
        lat_transect = 'Nan'
        lon_transect = 'Nan'

    earth_radius = 6371e3    # in meter
    try:
        dx = float(atr['X_STEP'])*np.pi/180.0*earth_radius * \
            np.sin(np.mean(lat)*np.pi/180)
        dy = float(atr['Y_STEP'])*np.pi/180.0*earth_radius
        DX = (x-x0)*dx
        DY = (y-y0)*dy
        D = np.hypot(DX, DY)
        print('geo coordinate:')
        print('profile length = ' + str(D[-1]/1000.0) + ' km')
        # df0_km=dist_point_from_line(mf,cf,x0,y0,dx,dy)
    except:
        dx = float(atr['RANGE_PIXEL_SIZE'])
        dy = float(atr['AZIMUTH_PIXEL_SIZE'])
        DX = (x-x0)*dx
        DY = (y-y0)*dy
        D = np.hypot(DX, DY)
        print('radar coordinate:')
        print('profile length = ' + str(D[-1]/1000.0) + ' km')
        # df0_km=dist_point_from_line(mf,cf,x0,y0,dx,dy)

    try:
        df0_km = dist_point_from_line(mf, cf, x0, y0, dx, dy)
    except:
        print('Fault line is not specified')

    transect = np.zeros([len(D), ntrans])
    transect[:, 0] = zi
    XX0 = []
    XX1 = []
    YY0 = []
    YY1 = []
    XX0.append(x0)
    XX1.append(x1)
    YY0.append(y0)
    YY1.append(y1)

    if ntrans > 1:
        m = float(y1-y0)/float(x1-x0)
        c = float(y0-m*x0)
        m1 = -1.0/m
        if lat_transect == 'Nan':
            for i in range(1, ntrans):
                X0 = i*dp/np.sqrt(1+m1**2)+x0
                Y0 = m1*(X0-x0)+y0
                X1 = i*dp/np.sqrt(1+m1**2)+x1
                Y1 = m1*(X1-x1)+y1
                zi = get_transect(z, X0, Y0, X1, Y1)
                transect[:, i] = zi
                XX0.append(X0)
                XX1.append(X1)
                YY0.append(Y0)
                YY1.append(Y1)
        else:
            transect_lat = np.zeros([len(D), ntrans])
            transect_lat[:, 0] = lat_transect
            transect_lon = np.zeros([len(D), ntrans])
            transect_lon[:, 0] = lon_transect

            for i in range(1, ntrans):
                X0 = i*dp/np.sqrt(1+m1**2)+x0
                Y0 = m1*(X0-x0)+y0
                X1 = i*dp/np.sqrt(1+m1**2)+x1
                Y1 = m1*(X1-x1)+y1
                zi = get_transect(z, X0, Y0, X1, Y1)
                lat_transect = get_transect(lat_all, X0, Y0, X1, Y1)
                lon_transect = get_transect(lon_all, X0, Y0, X1, Y1)
                transect[:, i] = zi
                transect_lat[:, i] = lat_transect
                transect_lon[:, i] = lon_transect
                XX0.append(X0)
                XX1.append(X1)
                YY0.append(Y0)
                YY1.append(Y1)

    #############################################
    try:
        m_prof_edge, c_prof_edge = line(XX0[0], YY0[0], XX0[-1], YY0[-1])
    except:
        print('Plotting one profile')

    ###############################################################################
    if flip_profile == 'yes':
        transect = np.flipud(transect)
        try:
            df0_km = np.max(D)-df0_km
        except:
            print('')

    print('******************************************************')
    try:
        gpsFile
    except:
        gpsFile = 'Nogps'
    print('GPS velocity file:')
    print(gpsFile)
    print('*******************************************************')
    if os.path.isfile(gpsFile):
        insarData = z
        del z
        fileName, fileExtension = os.path.splitext(gpsFile)
        #print fileExtension
        #if fileExtension =='.cmm4':
        #    print 'reading cmm4 velocities'
        #    Stations, gpsData = redGPSfile_cmm4(gpsFile)
        #    idxRef=Stations.index(refStation)
        #    Lon,Lat,Ve,Vn,Se,Sn,Corr,Hrate,H12=gpsData[idxRef,:]
        #    Lon=Lon-360.0
        #    Lat,Lon,Ve,Se,Vn,Sn,Corr,NumEpochs,timeSpan,AvgEpochTimes = gpsData[idxRef,:]
        #    Vu=0
        #else:
        #    Stations, gpsData = redGPSfile(gpsFile)
        #    idxRef=Stations.index(refStation)
        #    Lat,Lon,Vn,Ve,Sn,Se,Corr,Vu,Su = gpsData[idxRef,:]

        Stations, Lat, Lon, Ve, Se, Vn, Sn = readGPSfile(gpsFile, gps_source)
        idxRef = Stations.index(refStation)
        Length, Width = np.shape(insarData)
        #lat,lon,lat_step,lon_step = get_lat_lon(h5file,Length,Width)
        lat, lon, lat_step, lon_step, lat_all, lon_all = get_lat_lon(h5file)
        IDYref, IDXref = find_row_column(
            Lon[idxRef], Lat[idxRef], lon, lat, lon_step, lat_step)
        if (not np.isnan(IDYref)) and (not np.isnan(IDXref)):
            print('referencing InSAR data to the GPS station at : ' +
                  str(IDYref) + ' , ' + str(IDXref))
            if not np.isnan(insarData[IDYref][IDXref]):
                transect = transect - insarData[IDYref][IDXref]
                insarData = insarData - insarData[IDYref][IDXref]

            else:

                print("""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      WARNING: nan value for InSAR data at the refernce pixel!
               reference station should be a pixel with valid value in InSAR data.

               please select another GPS station as the reference station.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                   """)
                sys.exit(1)
        else:
            print('WARNING:')
            print('Reference GPS station is out of the area covered by InSAR data')
            print('please select another GPS station as the reference station.')
            sys.exit(1)

        try:
            stationsList
        except:
            stationsList = Stations

       # theta=23.0*np.pi/180.0
        if os.path.isfile(incidence_file):
            print('Using exact look angle for each pixel')
            h5file_theta = h5py.File(incidence_file, 'r')
            dset = h5file_theta['mask'].get('mask')
            theta = dset[0:dset.shape[0], 0:dset.shape[1]]
            theta = theta*np.pi/180.0
        else:
            print('Using average look angle')
            theta = np.ones(np.shape(insarData))*23.0*np.pi/180.0

        heading = 193.0*np.pi/180.0

        #unitVec=[-np.sin(theta)*np.sin(heading),-np.cos(heading)*np.sin(theta),-np.cos(theta)]
        unitVec = [np.cos(heading)*np.sin(theta), -np.sin(theta)
                   * np.sin(heading), 0]  # -np.cos(theta)]

        # [0.0806152480932643, 0.34918300221540616, -0.93358042649720174]
        # print unitVec
        # unitVec=[0.3,-0.09,0.9]
        # unitVec=[-0.3,0.09,-0.9]
        # unitVec=[-0.3,0.09,0]

        # print '*******************************************'
        # print 'unit vector to project GPS to InSAR LOS:'
        # print unitVec
        # print '*******************************************'
        # gpsLOS_ref=unitVec[0]*Ve[idxRef]+unitVec[1]*Vn[idxRef]#+unitVec[2]*Vu[idxRef]

        # print np.shape(theta)
        # print IDYref
        # print IDXref
        # print theta[IDYref,IDXref]

        gpsLOS_ref = gps_to_LOS(
            Ve[idxRef], Vn[idxRef], theta[IDYref, IDXref], heading)
        print('%%%%%%^^^^^^^%%%%%%%%')
        print(gpsLOS_ref/1000.0)
        # insarData=insarData -gpsLOS_ref/1000.0
        # transect = transect -gpsLOS_ref/1000.0

        GPS = []
        GPS_station = []
        GPSx = []
        GPSy = []
        GPS_lat = []
        GPS_lon = []
        for st in stationsList:
            try:
                idx = Stations.index(st)

                #gpsLOS = unitVec[0]*Ve[idx]+unitVec[1]*Vn[idx]#+unitVec[2]*Vu[idx]

                #gpsLOS = gps_to_LOS(Ve[idx],Vn[idx],theta[idx],heading)
                #gpsLOS=gpsLOS-gpsLOS_ref

                IDY, IDX = find_row_column(
                    Lon[idx], Lat[idx], lon, lat, lon_step, lat_step)
                print(theta[IDY, IDX])
                gpsLOS = gps_to_LOS(Ve[idx], Vn[idx], theta[IDY, IDX], heading)
                #gpsLOS = gpsLOS-gpsLOS_ref

                if which_gps == 'all':
                    if theta[IDY, IDX] != 0.0:
                        GPS.append(gpsLOS-gpsLOS_ref)
                        GPS_station.append(st)
                        GPSx.append(IDX)
                        GPSy.append(IDY)
                        GPS_lat.append(Lat[idx])
                        GPS_lon.append(Lon[idx])
                elif not np.isnan(insarData[IDY][IDX]):
                    if theta[IDY, IDX] != 0.0:
                        GPS.append(gpsLOS-gpsLOS_ref)
                        GPS_station.append(st)
                        GPSx.append(IDX)
                        GPSy.append(IDY)
                        GPS_lat.append(Lat[idx])
                        GPS_lon.append(Lon[idx])
            except:
                NoInSAR = 'yes'

        #print GPS_station
        #print gpsLOS
        DistGPS = []
        GPS_in_bound = []
        GPS_in_bound_st = []
        GPSxx = []
        GPSyy = []
        for i in range(len(GPS_station)):
            gx = GPSx[i]
            gy = GPSy[i]
            #print '******************'
            #print gx
            #print gy
            if which_gps in ['all', 'insar']:
                check_result = 'True'
            else:
                check_result = check_st_in_box(
                    gx, gy, x0, y0, x1, y1, X0, Y0, X1, Y1)

            if check_result == 'True':
                check_result2 = check_st_in_box2(
                    gx, gy, x0, y0, x1, y1, X0, Y0, X1, Y1)
                GPS_in_bound_st.append(GPS_station[i])
                GPS_in_bound.append(GPS[i])
                GPSxx.append(GPSx[i])
                GPSyy.append(GPSy[i])
            #gy=y0+1
            #gx=x0+1
            #gxp,gyp=get_intersect(m,c,gx,gy)
            #Dx=dx*(gx-gxp);Dy=dy*(gy-gyp)
            #print gxp
            #print gyp
                # distance of GPS station from the first profile line
                dg = dist_point_from_line(m, c, gx, gy, 1, 1)
            #DistGPS.append(np.hypot(Dx,Dy))
            #X0=dg/np.sqrt(1+m1**2)+x0
            #Y0=m1*(X0-x0)+y0
            #DistGPS.append(np.hypot(dx*(gx-X0), dy*(gy-Y0)))

                DistGPS.append(dist_point_from_line(
                    m_prof_edge, c_prof_edge, GPSx[i], GPSy[i], dx, dy))

        print('****************************************************')
        print('GPS stations in the profile area:')
        print(GPS_in_bound_st)
        print('****************************************************')
        GPS_in_bound = np.array(GPS_in_bound)
        DistGPS = np.array(DistGPS)
        #axes[1].plot(DistGPS/1000.0, -1*GPS_in_bound/1000, 'bo')

    if gpsFile == 'Nogps':

        insarData = z
        GPSxx = []
        GPSyy = []
        GPSx = []
        GPSy = []
        GPS = []
        XX0[0] = x0
        XX1[0] = x1
        YY0[0] = y0
        YY1[0] = y1

    # else:

    print('****************')
    print('flip up-down')
    print(flip_updown)

    if flip_updown == 'yes' and gpsFile != 'Nogps':
        print('Flipping up-down')
        transect = -1*transect
        GPS_in_bound = -1*GPS_in_bound
    elif flip_updown == 'yes':
        print('Flipping up-down')
        transect = -1*transect

    if flip_profile == 'yes' and gpsFile != 'Nogps':

        GPS = np.flipud(GPS)
        GPS_in_bound = np.flipud(GPS_in_bound)
        DistGPS = np.flipud(max(D)-DistGPS)

    fig, axes = plt.subplots(nrows=2)
    axes[0].imshow(insarData)
    for i in range(ntrans):
        axes[0].plot([XX0[i], XX1[i]], [YY0[i], YY1[i]], 'r-')

    axes[0].plot(GPSx, GPSy, 'b^')
    axes[0].plot(GPSxx, GPSyy, 'k^')
    if gpsFile != 'Nogps':
        axes[0].plot(IDXref, IDYref, 'r^')
    axes[0].axis('image')
    axes[1].plot(D/1000.0, transect, 'ko', ms=1)

    avgInSAR = np.array(nanmean(transect, axis=1))
    stdInSAR = np.array(nanstd(transect, axis=1))
    #print avgInSAR
    #print stdInSAR

    # std=np.std(transect,1)
    # axes[1].plot(D/1000.0, avgInSAR, 'r-')

    try:
        axes[1].plot(DistGPS/1000.0, -1*GPS_in_bound/1000, 'b^', ms=10)
    except:
        print('')
    # pl.fill_between(x, y-error, y+error,alpha=0.6, facecolor='0.20')
    # print transect
    #############################################################################

    fig2, axes2 = plt.subplots(nrows=1)
    axes2.imshow(insarData)
    # for i in range(ntrans):
    axes2.plot([XX0[0], XX1[0]], [YY0[0], YY1[0]], 'k-')
    axes2.plot([XX0[-1], XX1[-1]], [YY0[-1], YY1[-1]], 'k-')
    axes2.plot([XX0[0], XX0[-1]], [YY0[0], YY0[-1]], 'k-')
    axes2.plot([XX1[0], XX1[-1]], [YY1[0], YY1[-1]], 'k-')

    try:
        axes2.plot([Xf0, Xf1], [Yf0, Yf1], 'k-')
    except:
        FaultLine = 'None'

    axes2.plot(GPSx, GPSy, 'b^')
    axes2.plot(GPSxx, GPSyy, 'k^')
    if gpsFile != 'Nogps':
        axes2.plot(IDXref, IDYref, 'r^')
    axes2.axis('image')

    figName = 'transect_area.png'
    print('writing '+figName)
    plt.savefig(figName)

    #############################################################################
    fig = plt.figure()
    fig.set_size_inches(10, 4)
    ax = plt.Axes(fig, [0., 0., 1., 1.], )
    ax = fig.add_subplot(111)
    if display_InSAR in ['on', 'On', 'ON']:
        ax.plot(D/1000.0, transect*1000, 'o', ms=1, mfc='Black', linewidth='0')


    ############################################################################
    # save the profile data:
    if save_to_mat in ['ON', 'on', 'On', 'yes', 'y', 'YES', 'Yes']:
        matFile = 'transect.mat'
        dataset = {}
        dataset['datavec'] = transect
        try:
            dataset['lat'] = transect_lat
            dataset['lon'] = transect_lon
        except:
            dataset['lat'] = 'Nan'
            dataset['lon'] = 'Nan'
        dataset['Unit'] = 'm'
        dataset['Distance_along_profile'] = D
        print('*****************************************')
        print('')
        print('writing transect to >>> '+matFile)
        sio.savemat(matFile, {'dataset': dataset})
        print('')
        print('*****************************************')
    ############################################################################
    #ax.plot(D/1000.0, avgInSAR*1000, 'r-')

    #ax.plot(D/1000.0,transect*1000/(np.sin(23.*np.pi/180.)*np.cos(38.*np.pi/180.0)),'o',ms=1,mfc='Black', linewidth='0')
    #ax.plot(D/1000.0, avgInSAR*1000/(np.sin(23.*np.pi/180.)*np.cos(38.*np.pi/180.0)), 'r-')

    #############################################################################
    if disp_std in ['on', 'On', 'ON']:

        for i in np.arange(0.0, 1.01, 0.01):
            # ,color='#DCDCDC')#'LightGrey')
            ax.plot(D/1000.0, (avgInSAR-i*stdInSAR) *
                    1000, '-', color='#DCDCDC', alpha=0.5)
        for i in np.arange(0.0, 1.01, 0.01):
            ax.plot(D/1000.0, (avgInSAR+i*stdInSAR)*1000, '-',
                    color='#DCDCDC', alpha=0.5)  # 'LightGrey')
    #############################################################################
    if display_Average in ['on', 'On', 'ON']:
        ax.plot(D/1000.0, avgInSAR*1000, 'r-')
    ###########
    #ax.fill_between(D/1000.0,
    #                (avgInSAR-stdInSAR)*1000,
    #                (avgInSAR+stdInSAR)*1000,
    #                where=(avgInSAR+stdInSAR)*1000>=(avgInSAR-stdInSAR)*1000,
    #                alpha=1,
    #                facecolor='Red')

    try:
        ax.plot(DistGPS/1000.0, -1*GPS_in_bound, '^', ms=10, mfc='Cyan')
    except:
        print('')
    ax.set_ylabel('LOS velocity [mm/yr]', fontsize=26)
    ax.set_xlabel('Distance along profile [km]', fontsize=26)

    # print '******************'
    # print 'Dsitance of fault from the beginig of profile(km):'
    # print df0_km/1000.0

    ###################################################################
    # lower and higher bounds for diplaying the profile

    try:
        lbound
        hbound
    except:
        lbound = np.nanmin(transect)*1000
        hbound = np.nanmax(transect)*1000

    ###################################################################
    # To plot the Fault location on the profile
    try:
        ax.plot([df0_km/1000.0, df0_km/1000.0], [lbound, hbound],
                '--', color='black', linewidth='2')
    except:
        fault_loc = 'None'

    ###################################################################
    try:
        ax.set_ylim(lbound, hbound)
    except:
        ylim = 'no'

    # try:
    #      ax.set_xlim(-10,300)
    # except:
    #     xlim='no'

    ##########
    # Temporary To plot DEM
        # try:
    #    majorLocator = MultipleLocator(5)
    #    ax.yaxis.set_major_locator(majorLocator)
    #    minorLocator   = MultipleLocator(1)
    #    ax.yaxis.set_minor_locator(minorLocator)

    #    plt.tick_params(which='major', length=15,width=2)
    #    plt.tick_params(which='minor', length=6,width=2)

    #    try:
    #       for tick in ax.xaxis.get_major_ticks():
    #                tick.label.set_fontsize(26)
    #       for tick in ax.yaxis.get_major_ticks():
    #                tick.label.set_fontsize(26)
    #
    #       plt.tick_params(which='major', length=15,width=2)
    #       plt.tick_params(which='minor', length=6,width=2)
    #    except:
    #       print 'couldn not fix the ticks! '

    figName = 'transect.png'
    print('writing '+figName)
    plt.savefig(figName)
    print('')
    print('________________________________')
    plt.show()


#############################################################################
if __name__ == '__main__':
    main(sys.argv[1:])



#! /usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, 2017                               #
############################################################
# Recommended usage:
#   from mintpy.simulation import variance as var


import numpy as np
import pyproj

from mintpy.utils import ptime


def sample_data(lat, lon, mask=None, num_sample=500):
    ''''''
    ## Flatten input data
    lat = lat.flatten()
    lon = lon.flatten()

    ## Check number of samples and number of pixels
    num_pixel = len(lat)
    if num_sample > num_pixel:
        print('Number of samples > number of pixels, fix number of samples to number of pixels.')
        num_sample = num_pixel

    # Check input mask
    if mask is None:
        mask = np.ones((num_pixel), dtype=np.bool_)
    mask = mask.flatten()

    # Random select samples
    idx = np.arange(num_pixel)[mask]
    rng = np.random.default_rng()
    idx_sample = rng.choice(idx, size=int(num_sample))
    lat_sample = lat[idx_sample]
    lon_sample = lon[idx_sample]
    return idx_sample, lat_sample, lon_sample


def get_distance(lat, lon, i):
    '''Return the distance of all points in lat/lon from its ith point'''
    lat1 = lat[i]*np.ones(lat.shape)
    lon1 = lon[i]*np.ones(lon.shape)

    g = pyproj.Geod(ellps='WGS84')
    dist = g.inv(lon1, lat1, lon, lat)[2]
    return dist


def structure_function(data, lat, lon, step=5e3, min_pair_num=100e3, print_msg=True):
    num_sample = len(data)
    distance = np.zeros(num_sample**2)
    variance = np.zeros(num_sample**2)
    if print_msg:
        prog_bar = ptime.progressBar(maxValue=num_sample)
    for i in range(num_sample):
        distance[i*num_sample:(i+1)*num_sample] = get_distance(lat, lon, i)
        variance[i*num_sample:(i+1)*num_sample] = np.square(data - data[i])
        if print_msg:
            prog_bar.update(i+1, every=10)
    if print_msg:
        prog_bar.close()

    (bin_dist,
     bin_struct_func,
     bin_struct_func_std) = bin_variance(distance, variance,
                                         step=step,
                                         min_pair_num=min_pair_num,
                                         print_msg=print_msg)
    return bin_dist, bin_struct_func, bin_struct_func_std


def bin_variance(distance, variance, step=5e3, min_pair_num=100e3, print_msg=True):
    x_steps = np.arange(0,np.max(distance),step)
    num_step = len(x_steps)
    var = np.zeros(x_steps.shape)
    var_std = np.zeros(var.shape)
    p_num = np.zeros(x_steps.shape)

    if print_msg:
        prog_bar = ptime.progressBar(maxValue=num_step)
    for i in range(num_step):
        x = x_steps[i]
        idx = (distance > max(0, x-step/2.)) * (distance < x+step/2.)
        p_num[i] = np.sum(idx)
        var[i] = np.mean(variance[idx])
        var_std[i] = np.std(variance[idx])
        if print_msg:
            prog_bar.update(i+1, every=10)
    if print_msg:
        prog_bar.close()

    max_step_idx = int(max(np.argwhere(p_num > min_pair_num)))
    return x_steps[0:max_step_idx], var[0:max_step_idx], var_std[0:max_step_idx]



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, Antonio Valentino, Aug 2018        #
############################################################

import os
import sys

from mintpy.utils.arg_utils import create_argument_parser

###########################################################################################
EXAMPLE = """Example:
  remove_hdf5_dset.py  ifgramStack.h5  unwrapPhase_phaseClosure
  remove_hdf5_dset.py  ifgramStack.h5  unwrapPhase_phaseClosure  unwrapPhase_bridging
  remove_hdf5_dset.py  velocity.h5     velocityStd
"""


def create_parser(subparsers=None):
    synopsis = 'Remove an existing dataset from HDF5 file'
    epilog = EXAMPLE
    name = __name__.split('.')[-1]
    parser = create_argument_parser(
        name, synopsis=synopsis, description=synopsis, epilog=epilog, subparsers=subparsers)

    parser.add_argument('file', type=str, help='HDF5 file of interest')
    parser.add_argument('dset', type=str, nargs='+', help='dataset to be removed.')

    return parser


def cmd_line_parse(iargs=None):
    # parse
    parser = create_parser()
    inps = parser.parse_args(args=iargs)

    # check: input file extension
    if os.path.splitext(inps.file)[1] not in ['.h5', '.he5']:
        raise ValueError(f'input file is NOT HDF5: {inps.file}')

    return inps


###########################################################################################
def run_remove_hdf5_dset(fname, ds_names):
    """Remove a dataset from the given HDF5 file.

    Parameters: fname    - str, path to the HDF5 data file
                ds_names - list(str), name of the HDF5 dataset to be removed.
    """
    import h5py

    from mintpy.utils import writefile

    # grab exiting dataset list
    with h5py.File(fname, 'r') as f:
        dset_list = list(f.keys())

    # check if given dataset exists
    if any(i not in dset_list for i in ds_names):
        msg = f'input dataset ({ds_names}) do not exist!'
        msg += f'\nAvailable datasets: {dset_list}'
        raise ValueError(msg)

    # update file
    writefile.remove_hdf5_dataset(fname, ds_names, print_msg=True)
    print('Done.')

    return


###########################################################################################
def main(iargs=None):
    # parse
    inps = cmd_line_parse(iargs)

    # run
    run_remove_hdf5_dset(inps.file, inps.dset)


###########################################################################################
if __name__ == '__main__':
    main(sys.argv[1:])



############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, 2019                               #
############################################################


import os
import warnings  # suppress UserWarning from matplotlib

warnings.filterwarnings("ignore", category=UserWarning, module="matplotlib")

import matplotlib.pyplot as plt
import numpy as np
import scipy.io as sio

from mintpy.utils import writefile


##############################################################################
def grab_data_paths_from_inp_file(inp_file):
    """Grab data paths from inp file."""
    data_paths = []
    with open(inp_file) as f:
        lines = f.readlines()
        lines = [i.strip() for i in lines if not i.startswith('%')]
        lines = [i for i in lines if i]
        for line in lines:
            c = [i.strip() for i in line.strip().split('=', 1)]
            if (len(c) >= 2 and c[0] == 'insar{insarID}.dataPath'):
                mat_file = c[1].replace('\n','').split(";")[0].strip()
                mat_file = mat_file.replace("'","").replace('"','')
                data_paths.append(mat_file)
    return data_paths


def gbis_mat2hdf5(inv_mat_file, display=True):
    """Convert InSAR related GBIS inversion result .mat file into HDF5 file."""
    out_dir = os.path.dirname(inv_mat_file)
    out_files = []

    print(f'read mat file: {inv_mat_file}')
    mat = sio.loadmat(inv_mat_file, struct_as_record=False, squeeze_me=True)

    # when num_file == 1
    if isinstance(mat['insar'], sio.matlab.mio5_params.mat_struct):
        mat['insar'] = [mat['insar']]
        mat['insarPlot'] = [mat['insarPlot']]
    num_file = len(mat['insar'])
    print(f'number of output HDF5 file: {num_file}')

    # grab optimal model parameter
    parValue = mat['invResults'].optimalmodel
    parName = mat['invResults'].model.parName
    modelName = parName[0].split()[0]

    mDict = {}
    for i in range(len(parValue)):
        key = parName[i].replace(' ', '_')
        mDict[key] = parValue[i]
    mDict['DEFORMATION_MODEL'] = modelName

    if display:
        fig_size = [12, 3*num_file]
        fig, axs = plt.subplots(nrows=num_file, ncols=3, figsize=fig_size)
        axs = axs.reshape(-1,3)   #convert to 2D array when num_file is 1.
        print(f'creating figure in size of {fig_size}')

    for i in range(num_file):
        insar_mat_file = mat['insar'][i].dataPath
        if not os.path.isfile(insar_mat_file):
            inp_file = os.path.dirname(os.path.dirname(inv_mat_file))+'.inp'
            insar_mat_file = grab_data_paths_from_inp_file(inp_file)[i]
        print('-'*30)
        print(f'read mask from file: {insar_mat_file}')

        # read 1D height and 2D mask
        hgt1 = sio.loadmat(insar_mat_file, struct_as_record=False, squeeze_me=True)['Height']
        mask = sio.loadmat(insar_mat_file, struct_as_record=False, squeeze_me=True)['Mask']
        length, width = mask.shape

        # convert to 2D matrix
        insarPlot = mat['insarPlot'][i]
        out_file = os.path.join(out_dir, f'{insarPlot.name}.h5')
        out_files.append(out_file)

        hgt2 = np.zeros((length, width), dtype=np.float32) * np.nan
        data = np.zeros((length, width), dtype=np.float32) * np.nan
        model = np.zeros((length, width), dtype=np.float32) * np.nan
        residual = np.zeros((length, width), dtype=np.float32) * np.nan
        hgt2[mask!=0] = hgt1
        data[mask!=0] = insarPlot.data
        model[mask!=0] = insarPlot.model
        residual[mask!=0] = insarPlot.residual

        # prepare metadata
        meta = vars(sio.loadmat(insar_mat_file, struct_as_record=False, squeeze_me=True)['Metadata'])
        if '_fieldnames' in meta.keys():
            meta.pop('_fieldnames')
        meta['UNIT'] = 'm'
        meta['FILE_TYPE'] = 'displacement'
        meta['PROCESSOR'] = 'GBIS'
        for key, value in mDict.items():
            meta[key] = value

        # write to HDF5 file
        dsDict = {}
        dsDict['hgt'] = hgt2
        dsDict['data'] = data
        dsDict['model'] = model
        dsDict['residual'] = residual
        writefile.write(dsDict, out_file=out_file, metadata=meta)

        # plot
        if display:
            dlim = np.nanmax(np.abs(data))
            for ax, d in zip(axs[i, :], [data, model, residual]):
                im = ax.imshow(d, vmin=-dlim, vmax=dlim, cmap='jet')
                fig.colorbar(im, ax=ax)
            axs[i,0].set_ylabel('\n'.join(insarPlot.name.split('_', 1)))

    if display:
        print('showing ...')
        axs[0,0].set_title('data')
        axs[0,1].set_title('model')
        axs[0,2].set_title('residual')
        plt.show()
    return out_files



#! /usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, Dec 2019                           #
############################################################
# Recommended usage:
#   from mintpy.simulation import decorrelation as decor


import math

import numpy as np
import scipy.stats as stats
from matplotlib import pyplot as plt

from mintpy.utils import ptime


######################################## Statistics ############################################
def phase_pdf_ds(L, coherence=None, phi_num=1000, coh_step=0.005):
    """Marginal PDF of interferometric phase for distributed scatterers (DS)

    Eq. 66 (Tough et al., 1995) and Eq. 4.2.23 (Hanssen, 2001)

    Parameters: L         - int, number of independent looks
                coherence - 1D np.array for the range of coherence, with value < 1.0 for valid operation
                phi_num   - int, number of phase sample for the numerical calculation
                coh_step  - float, incremental step of coherence, used only if coherence is None.
    Returns:    pdf       - 2D np.array, phase PDF in size of (phi_num, len(coherence))
                coherence - 1D np.array for the range of coherence
    Example:
        coh = np.linspace(0.005, 1.0, 200) - 0.0025
        pdf, coh = phase_pdf_ds(1, coherence=coh)
    """

    def gamma(x):
        """
        Gamma function equivalent to scipy.special.gamma(x)
        :param x: float
        :return: float
        """
        # This function replaces scipy.special.gamma(x).
        # It is needed due to a bug where Dask workers throw an exception in which they cannot
        # find `scipy.special.gamma(x)` even when it is imported.

        # When the output of the gamma function is undefined, scipy.special.gamma(x) returns float('inf')
        # whereas math.gamma(x) throws an exception.
        try:
            return math.gamma(x)
        except ValueError:
            return float('inf')

    # default coherence value
    if coherence is None:
        coh_num = int(1. / coh_step)
        coherence = np.linspace(coh_step, 1., num=coh_num) - coh_step / 2.0

    coherence = np.array(coherence, np.float64).reshape(1, -1)
    phi = np.linspace(-np.pi, np.pi, phi_num, dtype=np.float64).reshape(-1, 1)

    # Phase PDF - Eq. 4.2.32 (Hanssen, 2001)
    A = np.power((1-np.square(coherence)), L) / (2*np.pi)
    A = np.tile(A, (phi_num, 1))
    B = gamma(2*L - 1) / ((gamma(L))**2 * 2**(2*(L-1)))

    beta = np.multiply(np.abs(coherence), np.cos(phi))
    C = np.divide((2*L - 1) * beta, np.power((1 - np.square(beta)), L+0.5))
    C = np.multiply(C, (np.pi/2 + np.arcsin(beta)))
    C += 1 / np.power((1 - np.square(beta)), L)

    sumD = 0
    if L > 1:
        for r in range(int(L)-1):
            D = gamma(L-0.5) / gamma(L-0.5-r)
            D *= gamma(L-1-r) / gamma(L-1)
            D *= (1 + (2*r+1)*np.square(beta)) / np.power((1 - np.square(beta)), r+2)
            sumD += D
        sumD /= (2*(L-1))

    pdf = B*C + sumD
    pdf = np.multiply(A, pdf)

    return pdf, coherence.astype(np.float32).flatten()


def phase_variance_ds(L,  coherence=None, coh_step=0.005):
    """Interferometric phase variance for distributed scatterers (DS)
    Eq. 2.1.2 (Box et al., 2015) and Eq. 4.2.27 (Hanssen, 2001)
    Parameters: L         - int, number of independent looks
                coherence - 1D np.array for the range of coherence, with value < 1.0 for valid operation
    Returns:    var       - 1D np.array, phase variance in size of (len(coherence)) in radians^2
                coherence - 1D np.array for the range of coherence
    Example:
        coh = np.linspace(0.005, 1.0, 200) - 0.0025
        var, coh = phase_variance_ds(1, coherence=coh)
    """

    # default coherence value
    if coherence is None:
        coh_num = int(1. / coh_step)
        coherence = np.linspace(coh_step, 1., num=coh_num) - coh_step / 2.0

    coherence = np.array(coherence, dtype=np.float64)
    phi_num = len(coherence)
    phi = np.linspace(-np.pi, np.pi, phi_num, dtype=np.float64).reshape(-1, 1)
    phi_step = 2*np.pi/phi_num

    pdf, coherence = phase_pdf_ds(L, coherence=coherence, phi_num=phi_num)
    var = np.sum(np.multiply(np.square(np.tile(phi, (1, len(coherence)))), pdf)*phi_step, axis=0)

    # assign negative value (when coherence is very close to 1, i.e. 0.999) to the min positive value
    flag = var <= 0
    if not np.all(flag):
        var[flag] = np.nanmin(var[~flag])
    else:
        var[flag] = np.finfo(np.float64).eps
    return var, coherence


def phase_variance_ps(L, coherence=None, epsilon=1e-3):
    """
    the Cramer-Rao bound (CRB) of phase variance
    Given by Eq. 25 (Rodriguez and Martin, 1992)and Eq 4.2.32 (Hanssen, 2001)
    Valid when coherence is close to 1.

    Parameters: L         - int, number of independent looks applied
                coherence - float / np.ndarray, spatial coherence
                epsilon   - float, a small number to avoid denominator to be zero
    Returns:    var       - float / np.ndarray, phase variance in radians^2
                coherence - float / np.ndarray, spatial coherence (for reference purpose)
    """
    if coherence is None:
        coherence = np.linspace(0.9, 1.-epsilon, 1000, dtype=np.float64)
    var = (1 - coherence**2) / (2 * int(L) * coherence**2)
    return var, coherence


def cross_correlation_std(N, coh, corr_type='intensity'):
    """Standard deviation of cross correlation for differential shift estimation.

    Reference:
        Bamler, R., and M. Eineder (2005), Accuracy of differential shift estimation by
            correlation and split-bandwidth interferometry for wideband and delta-k SAR systems,
            Geoscience and Remote Sensing Letters, IEEE, 2(2), 151-155, doi:10.1109/LGRS.2004.843203.
        De Zan, F. (2014), Accuracy of Incoherent Speckle Tracking for Circular Gaussian Signals,
            IEEE Geoscience and Remote Sensing Letters, 11(1), 264-267, doi:10.1109/LGRS.2013.2255259.

    Parameters: N         - int   / 2D np.ndarray in size of (n, 1), number of independent samples (resolution ceels)
                coh       - float / 2D np.ndarray in size of (1, c), spatial coherence
                corr_type - str, type of image samples used for cross correlation: complex or intensity.
    Returns:    std       - float / 2D np.ndarray in size of (n, c), standard deviation in the unit of pixel
    """
    if isinstance(N, np.ndarray):
        N = N.reshape(-1, 1)
    if isinstance(coh, np.ndarray):
        coh = coh.reshape(1, -1)

    if corr_type == 'complex':
        # equation (1) from Bamler and Eineder (2005)
        std = np.sqrt(3 / (2*N)) * np.sqrt(1 - coh**2) / (np.pi * coh)

    elif corr_type in ['intensity', 'amplitude']:
        # equation (13) from De Zan (2014)
        std = np.sqrt(3 / (10*N)) * np.sqrt(2 + 5*coh**2 - 7*coh**4) / (np.pi * coh**2)

    else:
        raise ValueError(f'un-recognized corr_type={corr_type}')

    return std


########################################## Simulations #########################################
def coherence2decorrelation_phase(coh, L, coh_step=0.01, num_repeat=1, scale=1.0, display=False, print_msg=True):
    """Simulate decorrelation phase based on coherence array/matrix
    based on the phase PDF of DS from Tough et al. (1995).

    Parameters: coh        - 1/2/3D np.ndarray of float32 for spatial coherence
                L          - int, number of independent looks
                coh_step   - float, step of coherence to generate lookup table
                num_repeat - int, number of repeatetion
    Returns:    pha        - np.ndarray of float32 for decorrelation phase in radians
                                for num_repeat == 1, pha.shape = coh.shape
                                for num_repeat >  1, pha.shape = (coh.size, num_repeat)
    """
    shape_orig = coh.shape
    coh = coh.reshape(-1,1)
    num_coh = coh.size

    # check number of looks
    L = int(L)
    msg = f'number of independent looks L={L}'
    if L > 80:
        L = 80
        msg += ', use L=80 to avoid dividing by 0 in calculation with negligible effect'
    if print_msg:
        print(msg)

    # code for debug
    debug_mode = False
    if debug_mode is True:
        decor = sample_decorrelation_phase(0.4, L, size=int(1e4), scale=scale)
        return decor

    # initiate output matrix
    pha = np.zeros((num_coh, num_repeat), dtype=np.float32)

    # sampling strategy
    num_step = int(1 / coh_step)
    if num_coh <= num_step * 2:
        # for small size --> loop through each input coherence
        for i in range(num_coh):
            pha[i,:] = sample_decorrelation_phase(coh[i], L, size=num_repeat, scale=scale)

    else:
        # for large size --> loop through coherence lookup table and map into input coherence matrix
        prog_bar = ptime.progressBar(maxValue=num_step, print_msg=print_msg)
        for i in range(num_step):
            # find index within the coherence intervals
            coh_i = i * coh_step
            flag = np.multiply(coh >= coh_i, coh < coh_i + coh_step).flatten()
            num_coh_i = np.sum(flag)
            if num_coh_i > 0:
                pha_i = sample_decorrelation_phase(coh_i, L, size=num_coh_i*num_repeat, scale=scale)
                pha[flag,:] = pha_i.reshape(-1, num_repeat)
            prog_bar.update(i+1, suffix=f'{coh_i:.3f}')
        prog_bar.close()

    if num_repeat == 1:
        pha = pha.reshape(shape_orig)

    # plot
    if display and len(shape_orig) == 2:
        fig, axs = plt.subplots(nrows=1, ncols=2, figsize=[6,3])
        axs[0].imshow(coh.reshape(shape_orig), vmin=0, vmax=1, cmap='gray')
        axs[1].imshow(pha[:,0].reshape(shape_orig), vmin=-np.pi, vmax=np.pi, cmap='jet')
        plt.show()

    return pha


def sample_decorrelation_phase(coherence, L, size=1, phi_num=1000, display=False, scale=1.0, font_size=12):
    '''Sample decorrelation phase based on PDF determined by L and coherence value
    Parameters: coherence - float, spatial coherence
                L         - int, number of independent looks
                phi_num   - int, sample number
    Returns:    phase     - 1D np.array in size of (size,), sampled phase in radians
    Examples:
        decor_noise = sample_decorrelation_phase(0.7, L=1, size=1e4, display=True)
    '''
    size = int(size)
    phiMax = np.pi * float(scale)

    # numerical solution of phase PDF for distributed scatterers
    pdf = phase_pdf_ds(int(L), coherence, phi_num=phi_num)[0].flatten()

    # generate phase distribution
    phi = np.linspace(-phiMax, phiMax, phi_num+1, endpoint=True)
    phi_dist = stats.rv_histogram((pdf, phi))

    # sample from the distribution
    phase = phi_dist.rvs(size=size)

    if display:
        _, ax = plt.subplots(figsize=[5,3])
        ax.hist(phase, bins=50, density=True, label='Sample\nHistogram\n(norm)')
        ax.plot(phi, phi_dist.pdf(phi), label='PDF')
        ax.plot(phi, phi_dist.cdf(phi), label='CDF')
        ax.set_xlabel('Phase', fontsize=font_size)
        ax.set_ylabel('Probability', fontsize=font_size)
        ax.set_title(r'L = %d, $\gamma$ = %.2f, sample size = %d' % (L, coherence, size), fontsize=font_size)
        ax.set_xlim([-np.pi, np.pi])
        ax.set_xticks([-np.pi, 0, np.pi])
        ax.set_xticklabels([r'-$\pi$', '0', r'$\pi$'], fontsize=font_size)
        ax.tick_params(direction='in', labelsize=font_size)
        ax.legend(fontsize=font_size)
        plt.savefig('DecorNoiseSampling.jpg', bbox_inches='tight', dpi=600)
        plt.show()
    return phase


#################################### Weight Functions #####################################
def coherence2phase_variance(coherence, L=32, scatter='DS', coh_step=0.005, print_msg=False):
    """Convert coherence to phase variance, depending on the type of scatterers

    For DS, it is equation (66) from Tough et al. (1995).
    For PS, it is equation (25) from Rodriguez and Martin (1992).

    Parameters: coherence - 1/2/3D np.ndarray of float32 for spatial coherence
                L         - int, number of independent looks
                scatter   - str, type of scatterers, PS or DS
    Returns:    variance  - 1/2/3D np.ndarray of float32 for the phase variance in radians^2
    """
    lineStr = f'    number of independent looks L={L}'
    if L > 80:
        L = 80
        lineStr += ', use L=80 to avoid dividing by 0 with negligible effect'
    if print_msg:
        print(lineStr)

    coh_num = int(1. / coh_step)
    coh_lut = np.linspace(coh_step, 1.0, num=coh_num) - coh_step / 2.0
    coh_min = np.min(coh_lut)
    coh_max = np.max(coh_lut)

    coherence = np.array(coherence)
    coherence[coherence < coh_min] = coh_min
    coherence[coherence > coh_max] = coh_max
    coherence_idx = np.array((coherence - coh_min) / coh_step, np.int16)

    scatter = scatter.upper()
    if scatter == 'DS':
        var_lut = phase_variance_ds(int(L), coh_lut)[0]
        variance = var_lut[coherence_idx]

    elif scatter == 'PS':
        variance = phase_variance_ps(int(L), coherence)[0]

    else:
        raise ValueError(f'un-recognized scatterer type: {scatter}')

    return variance


def coherence2fisher_info_index(data, L=32, epsilon=1e-3):
    """Convert coherence to Fisher information index (Seymour & Cumming, 1994, IGARSS)"""
    print(f'    number of independent looks L={L}')
    # prepare input data
    if data.dtype != np.float64:
        data = np.array(data, np.float64)
    data[data > 1-epsilon] = 1-epsilon
    # calculation
    data = 2.0 * int(L) * np.square(data) / (1 - np.square(data))
    return data


def coherence2weight(coh_data, weight_func='var', L=20, epsilon=5e-2, print_msg=True):
    """Convert spaital coherence to weight given the weight function name"""
    coh_data[np.isnan(coh_data)] = epsilon
    coh_data[coh_data < epsilon] = epsilon
    coh_data = np.array(coh_data, np.float64)

    # make sure L >= 1
    L = max(L, 1)

    # Calculate Weight matrix
    weight_func = weight_func.lower()
    if 'var' in weight_func:
        if print_msg:
            print('convert coherence to weight using inverse of phase variance')
            print('    with phase PDF for distributed scatterers from Tough et al. (1995)')
        weight = 1.0 / coherence2phase_variance(coh_data, L, print_msg=print_msg)

    elif any(i in weight_func for i in ['coh', 'lin']):
        if print_msg:
            print('use coherence as weight directly (Perissin & Wang, 2012; Tong et al., 2016)')
        weight = coh_data

    elif any(i in weight_func for i in ['fim', 'fisher']):
        if print_msg:
            print('convert coherence to weight using Fisher Information Index (Seymour & Cumming, 1994)')
        weight = coherence2fisher_info_index(coh_data, L)

    elif weight_func in ['no', 'sbas', 'uniform']:
        weight = None

    else:
        raise Exception('Un-recognized weight function: %s' % weight_func)

    # weight data is calculated in float64 for numerical precision
    # and saved in float32 after calculation for computing speed efficiency.
    if weight is not None:
        weight = np.array(weight, np.float32)
    del coh_data

    return weight



############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, Heresh Fattahi, 2013               #
############################################################


import os
import warnings

import numpy as np
from matplotlib import pyplot as plt

from mintpy.objects import ifgramStack
from mintpy.utils import plot as pp, readfile, utils as ut

# suppress UserWarning from matplotlib
warnings.filterwarnings("ignore", category=UserWarning, module="matplotlib")


###########################  Sub Function  #############################
def read_network_info(inps):
    ext = os.path.splitext(inps.file)[1]
    print('read temporal/spatial baseline info from file:', inps.file)

    ## 1. Read date, pbase, date12 and coherence
    if ext == '.h5':
        stack_obj = ifgramStack(inps.file)
        stack_obj.open()
        inps.date12List = stack_obj.get_date12_list(dropIfgram=False)
        inps.dateList = stack_obj.get_date_list(dropIfgram=False)
        inps.pbaseList = stack_obj.get_perp_baseline_timeseries(dropIfgram=False)

        if inps.dsetName in readfile.get_dataset_list(inps.file):
            inps.cohList = ut.spatial_average(
                inps.file,
                datasetName=inps.dsetName,
                maskFile=inps.maskFile,
                saveList=True,
                checkAoi=False,
            )[0]
        elif inps.dsetName == 'pbase':
            inps.cohList = np.abs(stack_obj.pbaseIfgram).tolist()

        elif inps.dsetName == 'tbase':
            inps.cohList = stack_obj.tbaseIfgram.tolist()

        else:
            inps.cohList = None
            msg = f'{inps.dsetName} NOT found in file: {inps.file}! '
            msg += 'Disable the color coding and continue'
            print(msg)

    elif ext == '.txt':
        inps.date12List = np.loadtxt(inps.file, dtype=bytes).astype(str)[:,0].tolist()

        # date12List --> dateList
        mDates = [i.split('_')[0] for i in inps.date12List]
        sDates = [i.split('_')[1] for i in inps.date12List]
        inps.dateList = sorted(list(set(mDates + sDates)))

        # pbase12List + date12List --> pbaseList
        pbase12List = np.loadtxt(inps.file, dtype=bytes).astype(float)[:,3]
        A = ifgramStack.get_design_matrix4timeseries(inps.date12List)[0]
        inps.pbaseList = np.zeros(len(inps.dateList), dtype=np.float32)
        inps.pbaseList[1:] = np.linalg.lstsq(A, np.array(pbase12List), rcond=None)[0]

        # cohList
        inps.cohList = np.loadtxt(inps.file, dtype=bytes).astype(float)[:,1]
    else:
        raise ValueError('un-recognized input file extention:', ext)
    print(f'number of acquisitions: {len(inps.dateList)}')
    print(f'number of interferograms: {len(inps.date12List)}')

    print(f'shift all perp baseline by {np.mean(inps.pbaseList)} to zero mean for plotting')
    inps.pbaseList -= np.mean(inps.pbaseList)

    # Optional: Read dropped date12 / date
    inps.dateList_drop = []
    inps.date12List_drop = []
    if ext == '.h5':
        inps.date12List_keep = ifgramStack(inps.file).get_date12_list(dropIfgram=True)
        inps.date12List_drop = sorted(list(set(inps.date12List) - set(inps.date12List_keep)))
        print('-'*50)
        print(f'number of interferograms marked as drop: {len(inps.date12List_drop)}')
        print(f'number of interferograms marked as keep: {len(inps.date12List_keep)}')

        mDates = [i.split('_')[0] for i in inps.date12List_keep]
        sDates = [i.split('_')[1] for i in inps.date12List_keep]
        inps.dateList_keep = sorted(list(set(mDates + sDates)))
        inps.dateList_drop = sorted(list(set(inps.dateList) - set(inps.dateList_keep)))
        print(f'number of acquisitions marked as drop: {len(inps.dateList_drop)}')
        if len(inps.dateList_drop) > 0:
            print(inps.dateList_drop)

    return inps


def check_colormap(inps):
    """Check input colormaps."""
    # adjust color jump if no --cmap-vlist is input
    if inps.cohList is not None and '--cmap-vlist' not in inps.argv:
        # use template value
        key = 'mintpy.network.minCoherence'
        if key in inps.template.keys():
            inps.cmap_vlist[1] = max(0.01, float(inps.template[key]))

        # calculate from dropped interferograms, if there is any
        elif len(inps.date12List_drop) > 0:
            idx_drop = [inps.date12List.index(i) for i in inps.date12List_drop]
            coh_drop = [inps.cohList[i] for i in idx_drop]
            inps.cmap_vlist[1] = max(coh_drop)
            print(f'max coherence of excluded interferograms: {max(coh_drop)}')

        # update default cmap_vlist[0] value if the manually input cmap_vlist[1] is too small
        if inps.cmap_vlist[1] <= inps.cmap_vlist[0]:
            inps.cmap_vlist[0] = min(0, inps.cmap_vlist[1])

    # in case the manually input list is not in order
    inps.cmap_vlist = sorted(inps.cmap_vlist)
    inps.colormap = pp.ColormapExt(inps.cmap_name, vlist=inps.cmap_vlist).colormap

    return inps


########################################################################
def plot_network(inps):
    """Plot all the network info."""

    # matplotlib backend setting
    if not inps.disp_fig:
        plt.switch_backend('Agg')

    ##---------- read / calculate
    inps = read_network_info(inps)

    ##---------- plot settings
    # color maps
    inps = check_colormap(inps)

    # figure size
    if not inps.fig_size:
        num_date = len(inps.dateList)
        if   num_date <= 100:  inps.fig_size = [ 6, 4]
        elif num_date <= 200:  inps.fig_size = [ 8, 4]
        elif num_date <= 300:  inps.fig_size = [10, 4]
        else:                  inps.fig_size = [12, 4]

    # save figure
    kwargs = dict(bbox_inches='tight', transparent=True, dpi=inps.fig_dpi)

    # labels in y-axis and colorbar
    inps.ds_name, inps.cbar_label = {
        'coherence' : ['Coherence',     'Average Spatial Coherence'],
        'offsetSNR' : ['SNR',           'Average Spatial SNR'],
        'tbase'     : ['Temp Baseline', 'Temp Baseline [day]'],
        'pbase'     : ['Perp Baseline', 'Perp Baseline [m]'],
    }[inps.dsetName]

    # figure names
    ext = 'Ion.pdf' if os.path.basename(inps.file).startswith('ion') else '.pdf'
    fig_names = {
        'coherence' : [i+ext for i in ['pbaseHistory',  'coherenceHistory', 'coherenceMatrix', 'network']],
        'offsetSNR' : [i+ext for i in ['pbaseHistory',        'SNRHistory',       'SNRMatrix', 'network']],
        'tbase'     : [i+ext for i in ['pbaseHistory',      'tbaseHistory',     'tbaseMatrix', 'network']],
        'pbase'     : [i+ext for i in ['pbaseHistory', 'pbaseRangeHistory',     'pbaseMatrix', 'network']],
    }[inps.dsetName]

    ##---------- plot
    # Fig 1 - Baseline History
    fig, ax = plt.subplots(figsize=inps.fig_size)
    ax = pp.plot_perp_baseline_hist(
        ax,
        inps.dateList,
        inps.pbaseList,
        vars(inps),
        inps.dateList_drop,
    )
    if inps.save_fig:
        fig.savefig(fig_names[0], **kwargs)
        print(f'save figure to {fig_names[0]}')

    if inps.cohList is not None:
        # Fig 2 - Min/Max Coherence History
        fig, ax = plt.subplots(figsize=inps.fig_size)
        ax = pp.plot_coherence_history(
            ax,
            inps.date12List,
            inps.cohList,
            p_dict=vars(inps),
        )
        if inps.save_fig:
            fig.savefig(fig_names[1], **kwargs)
            print(f'save figure to {fig_names[2]}')

        # Fig 3 - Coherence Matrix
        fig, ax = plt.subplots(figsize=[max(inps.fig_size), max(inps.fig_size)])
        ax = pp.plot_coherence_matrix(
            ax,
            inps.date12List,
            inps.cohList,
            inps.date12List_drop,
            p_dict=vars(inps),
        )[0]
        if inps.save_fig:
            fig.savefig(fig_names[2], **kwargs)
            print(f'save figure to {fig_names[1]}')

    # Fig 4 - Interferogram Network
    fig, ax = plt.subplots(figsize=inps.fig_size)
    ax = pp.plot_network(
        ax,
        inps.date12List,
        inps.dateList,
        inps.pbaseList,
        vars(inps),
        inps.date12List_drop,
    )
    if inps.save_fig:
        fig.savefig(fig_names[3], **kwargs)
        print(f'save figure to {fig_names[3]}')

    if inps.disp_fig:
        print('showing ...')
        plt.show()
    else:
        plt.close()

    return



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Antonio Valentino, Zhang Yunjun, Aug 2022        #
############################################################


import math
import sys

from mintpy.utils.arg_utils import create_argument_parser

############################################################
EXAMPLE = """example:
  mask.py  velocity.h5     -m Mask.h5
  mask.py  timeseries.h5   -m temporalCoherence.h5  -t 0.7
  mask.py  ifgramStack.h5  -m 100102_101120.cor     -t 0.9  -y  200 300  -x 300 400

  mask.py  filt_20060924_20090214.int -m waterMask.h5 -o filt_20060924_20090214_msk.int
  mask.py  filt_20060924_20090214.cor -m waterMask.h5 -o filt_20060924_20090214_msk.cor
"""

def create_parser(subparsers=None):
    synopsis = 'Mask file'
    epilog = EXAMPLE
    name = __name__.split('.')[-1]
    parser = create_argument_parser(
        name, synopsis=synopsis, description=synopsis, epilog=epilog, subparsers=subparsers)

    parser.add_argument('file', help='File to be masked')
    parser.add_argument('-m', '--mask', dest='mask_file', required=True,
                        help='mask for pixels used in ramp estimation')
    parser.add_argument('-o', '--outfile', help='Output file name.')

    # modify input mask
    parser.add_argument('-t', dest='threshold', type=float,
                        help='threshold value used for masking.\n' +
                        'if not specified, only pixels with mask value equal to zero is masked out.')
    parser.add_argument('--fill', dest='fill_value', type=float, default=math.nan,
                        help="fill masked out area with input value. i.e. \n"
                             "np.nan (default), 0, 1000, ... \n"
                             "If np.nan and input data matrix is not float/complex, convert matrix data type to np.float32.")
    parser.add_argument('-x', dest='subset_x', type=int, nargs=2,
                        help='subset range in x/cross-track/column direction')
    parser.add_argument('-y', dest='subset_y', type=int, nargs=2,
                        help='subset range in y/along-track/row direction')
    return parser


def cmd_line_parse(iargs=None):
    parser = create_parser()
    inps = parser.parse_args(args=iargs)
    return inps


############################################################
def main(iargs=None):
    # parse
    inps = cmd_line_parse(iargs)

    # import
    from mintpy.mask import mask_file

    # run
    mask_file(
        inps.file,
        mask_file=inps.mask_file,
        out_file=inps.outfile,
        fill_value=inps.fill_value,
        inps=inps,
    )


############################################################
if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3
# Author: Zhang Yunjun, Feb 2022
"""Test ascending/descending LOS decomposition into horizontal/vertical."""

import argparse
import os
import sys

import numpy as np
from matplotlib import pyplot as plt

from mintpy.asc_desc2horz_vert import asc_desc2horz_vert
from mintpy.utils import utils as ut

plt.rcParams.update({'font.size': 12})


################################################################################
EXAMPLE = """example:
  $MINTPY_HOME/tests/asc_desc2horz_vert.py
  $MINTPY_HOME/tests/asc_desc2horz_vert.py --plot
"""

def cmd_line_parse(iargs=None):
    # create parser
    parser = argparse.ArgumentParser(description='Test asc_desc2horz_vert.py',
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     epilog=EXAMPLE)
    parser.add_argument('--plot', dest='plot', action='store_true', help='Plot testing results.')

    # parsing
    inps = parser.parse_args(args=iargs)

    return inps


################################################################################
def main(iargs=None):

    inps = cmd_line_parse(iargs)
    print('-'*50)
    print(os.path.abspath(__file__))

    ## Setup
    # LOS incidence / azimuth angles
    los_inc_angle = np.array([ 30,   30], dtype=np.float32)
    los_az_angle  = np.array([102, -102], dtype=np.float32)

    # specify horz / vert [truth]
    length, width = 5, 5
    simH = np.ones((length, width), dtype=np.float32) * 0.5
    simV = np.ones((length, width), dtype=np.float32) * 1.0
    # azimuth angle in horizontal direction in degrees
    # measured from the north with anti-clockwise as positive
    # [0 for north, -90 for east]
    horz_az_angle = 30


    ## Testing
    # horz / vert --> east / north / up
    dE = simH * np.sin(np.deg2rad(horz_az_angle)) * -1
    dN = simH * np.cos(np.deg2rad(horz_az_angle))
    dU = simV

    # east / north / up --> asc / desc LOS
    dlos0 = ut.enu2los(dE, dN, dU, inc_angle=los_inc_angle[0], az_angle=los_az_angle[0])
    dlos1 = ut.enu2los(dE, dN, dU, inc_angle=los_inc_angle[1], az_angle=los_az_angle[1])

    # asc / desc LOS --> horz / vert [estimation]
    dlos = np.vstack((dlos0.reshape(1, length, width), dlos1.reshape(1, length, width)))
    estH, estV = asc_desc2horz_vert(dlos, los_inc_angle, los_az_angle, horz_az_angle)

    # check difference between the trush and estimation
    print(f'mean difference for horz / vert: {np.nanmean(estH - simH)} / {np.nanmean(estH - simH)}')
    assert np.allclose(simH, estH)
    assert np.allclose(simV, estV)


    # Plotting
    if inps.plot:
        print(f'plot test result of {os.path.basename(__file__)}')
        fig, axs = plt.subplots(nrows=3, ncols=4, figsize=[8, 6], sharex=True, sharey=True)
        kwargs = dict(vmin=-1.5, vmax=1.5, cmap='RdBu', interpolation='nearest')

        # horz / vert [truth]
        ax = axs[0, 0];  im = ax.imshow(simH,  **kwargs);  ax.set_title('Horz [sim]')
        ax = axs[1, 0];  im = ax.imshow(simV,  **kwargs);  ax.set_title('Vert [sim]')
        ax = axs[2, 0];  ax.axis('off')
        # east / north / up
        ax = axs[0, 1];  im = ax.imshow(dE,    **kwargs);  ax.set_title('East [sim]')
        ax = axs[1, 1];  im = ax.imshow(dN,    **kwargs);  ax.set_title('North [sim]')
        ax = axs[2, 1];  im = ax.imshow(dU,    **kwargs);  ax.set_title('Up [sim]')
        # asc / desc
        ax = axs[0, 2];  im = ax.imshow(dlos0, **kwargs);  ax.set_title('Asc [obs]')
        ax = axs[1, 2];  im = ax.imshow(dlos1, **kwargs);  ax.set_title('Desc [obs]')
        ax = axs[2, 2];  ax.axis('off')
        # horz / vert [estimation]
        ax = axs[0, 3];  im = ax.imshow(estH,  **kwargs);  ax.set_title('Horz [est]')
        ax = axs[1, 3];  im = ax.imshow(estV,  **kwargs);  ax.set_title('Vert [est]')
        ax = axs[2, 3];  ax.axis('off')

        # axis format
        fig.tight_layout()
        # colorbar
        cax = fig.add_axes([0.6, 0.2, 0.3, 0.02])
        fig.colorbar(im, cax=cax, orientation='horizontal')
        plt.show()


################################################################################
if __name__ == '__main__':
    main(sys.argv[1:])



############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, Nov 2018                           #
############################################################


import os

import matplotlib.pyplot as plt
import numpy as np

from mintpy import view
from mintpy.objects import ifgramStack
from mintpy.utils import plot as pp, readfile, utils as ut


###########################  Sub Function  #############################
def read_network_info(inps):
    """Read the network information"""
    ftype = readfile.read_attribute(inps.ifgram_file)['FILE_TYPE']
    if ftype != 'ifgramStack':
        raise ValueError(f'input file {inps.ifgram_file} is not ifgramStack: {ftype}')

    obj = ifgramStack(inps.ifgram_file)
    obj.open(print_msg=inps.print_msg)
    inps.date12_list = obj.get_date12_list(dropIfgram=False)
    date12_kept = obj.get_date12_list(dropIfgram=True)
    inps.ex_date12_list = sorted(list(set(inps.date12_list) - set(date12_kept)))
    inps.date_list = obj.get_date_list(dropIfgram=False)
    vprint(f'number of all     interferograms: {len(inps.date12_list)}')
    vprint(f'number of dropped interferograms: {len(inps.ex_date12_list)}')
    vprint(f'number of kept    interferograms: {len(inps.date12_list) - len(inps.ex_date12_list)}')
    vprint(f'number of acquisitions: {len(inps.date_list)}')

    if inps.lalo:
        if not inps.lookup_file:
            lookup_file = os.path.join(os.path.dirname(inps.ifgram_file), 'geometry*.h5')
            inps.lookup_file = ut.get_lookup_file(filePattern=lookup_file)
        coord = ut.coordinate(obj.metadata, lookup_file=inps.lookup_file)
        inps.yx = coord.geo2radar(inps.lalo[0], inps.lalo[1])[0:2]

    if not inps.yx:
        inps.yx = (obj.refY, obj.refX)
        vprint(f'plot initial coherence matrix at reference pixel: {inps.yx}')
    return inps


class coherenceMatrixViewer():
    """class for plot_coherence_matrix
    Example:
        from mintpy.plot_coherence_matrix import coherenceMatrixViewer
        cmd = 'plot_coherence_matrix.py ./inputs/ifgramStack.h5 --noverbose --figsize 9 3 --yx 216 310'
        obj = coherenceMatrixViewer(cmd)
        obj.configure()
        obj.plot()
    """
    def __init__(self, cmd=None, iargs=None):
        if cmd:
            iargs = cmd.split()[1:]
        self.cmd = cmd
        self.iargs = iargs

        # figure variables
        self.figname = 'Coherence matrix'
        self.fig_size = None
        self.fig = None
        self.ax_img = None
        self.ax_mat = None
        return

    def configure(self, inps):
        global vprint
        vprint = print if inps.print_msg else lambda *args, **kwargs: None

        # matplotlib backend setting
        if not inps.disp_fig:
            plt.switch_backend('Agg')

        # read network info
        inps = read_network_info(inps)
        # copy inps to self object
        for key, value in inps.__dict__.items():
            setattr(self, key, value)

        # auto figure size
        if not self.fig_size:
            ds_shape = readfile.read(self.img_file)[0].shape
            fig_size = pp.auto_figure_size(ds_shape, disp_cbar=True, scale=0.7)
            self.fig_size = [fig_size[0]+fig_size[1], fig_size[1]]
            vprint(f'create figure in size of {self.fig_size} inches')

        # read aux data
        # 1. temporal coherence value
        self.tcoh = None
        if self.tcoh_file:
            self.tcoh = readfile.read(self.tcoh_file)[0]
        # 2. minimum used coherence from template file
        self.min_coh_used = 0.0
        if self.template_file:
            template = readfile.read_template(self.template_file)
            template = ut.check_template_auto_value(template)
            if template['mintpy.networkInversion.maskDataset'] == 'coherence':
                self.min_coh_used = float(template['mintpy.networkInversion.maskThreshold'])
                vprint('Pixel-wised masking is applied in invert_network step')
        return

    def plot(self):
        # Figure 1
        self.fig = plt.figure(self.figname, figsize=self.fig_size)

        # Axes 1 - Image
        self.ax_img = self.fig.add_axes([0.05, 0.1, 0.4, 0.8])
        view_cmd = self.view_cmd.format(self.img_file)
        d_img, atr, inps_img = view.prep_slice(view_cmd)
        if all(i is not None for i in self.yx):
            inps_img.pts_marker = 'r^'
            inps_img.pts_yx = np.array(self.yx).reshape(-1, 2)

            # point yx --> lalo for geocoded product
            if 'Y_FIRST' in atr.keys():
                coord = ut.coordinate(atr)
                inps_img.pts_lalo = np.array(coord.radar2geo(self.yx[0], self.yx[1])[0:2]).reshape(-1,2)

        inps_img.print_msg = self.print_msg
        self.ax_img = view.plot_slice(self.ax_img, d_img, atr, inps_img)[0]

        # coordinate info
        self.coord = ut.coordinate(atr)
        self.fig_coord = inps_img.fig_coord

        # Axes 2 - coherence matrix
        self.ax_mat = self.fig.add_axes([0.55, 0.125, 0.40, 0.75])
        self.colormap = pp.ColormapExt(self.cmap_name, vlist=self.cmap_vlist).colormap
        if all(i is not None for i in self.yx):
            self.plot_coherence_matrix4pixel(self.yx)

        # Link the canvas to the plots.
        self.cid = self.fig.canvas.mpl_connect('button_press_event', self.update_coherence_matrix)
        if self.disp_fig:
            plt.show()
        return

    def plot_coherence_matrix4pixel(self, yx):
        """Plot coherence matrix for one pixel
        Parameters: yx : list of 2 int
        """
        self.ax_mat.cla()

        # read coherence
        box = (yx[1], yx[0], yx[1]+1, yx[0]+1)
        coh = readfile.read(self.ifgram_file, datasetName='coherence', box=box)[0]

        # ex_date for pixel-wise masking during network inversion
        ex_date12_list = self.ex_date12_list[:]   #local copy
        if self.min_coh_used > 0.:
            ex_date12_list += np.array(self.date12_list)[coh < self.min_coh_used].tolist()
            ex_date12_list = sorted(list(set(ex_date12_list)))

        # prep metadata
        plotDict = {}
        plotDict['fig_title'] = f'Y = {yx[0]}, X = {yx[1]}'
        # display temporal coherence value of the pixel
        if self.tcoh_file:
            tcoh = self.tcoh[yx[0], yx[1]]
            plotDict['fig_title'] += f', tcoh = {tcoh:.2f}'
        plotDict['colormap'] = self.colormap
        plotDict['cmap_vlist'] = self.cmap_vlist
        plotDict['disp_legend'] = False

        # plot
        coh_mat = pp.plot_coherence_matrix(
            self.ax_mat,
            date12List=self.date12_list,
            cohList=coh.tolist(),
            date12List_drop=ex_date12_list,
            p_dict=plotDict)[1]

        self.ax_mat.annotate('ifgrams\navailable', xy=(0.05, 0.05), xycoords='axes fraction', fontsize=12)
        self.ax_mat.annotate('ifgrams\nused', ha='right', xy=(0.95, 0.85), xycoords='axes fraction', fontsize=12)

        # status bar
        def format_coord(x, y):
            row, col = int(y+0.5), int(x+0.5)
            date12 = sorted([self.date_list[row], self.date_list[col]])
            date12 = [f'{i[0:4]}-{i[4:6]}-{i[6:8]}' for i in date12]
            return f'x={date12[0]}, y={date12[1]}, v={coh_mat[row, col]:.3f}'
        self.ax_mat.format_coord = format_coord

        # info
        msg = f'pixel in yx = {tuple(yx)}, '
        msg += f'min/max spatial coherence: {np.min(coh):.2f} / {np.max(coh):.2f}, '
        if self.tcoh_file:
            msg += f'temporal coherence: {tcoh:.2f}'
        vprint(msg)

        # update figure
        self.fig.canvas.draw_idle()
        self.fig.canvas.flush_events()
        return

    def update_coherence_matrix(self, event):
        if event.inaxes == self.ax_img:
            if self.fig_coord == 'geo':
                yx = [self.coord.lalo2yx(event.ydata, coord_type='lat'),
                      self.coord.lalo2yx(event.xdata, coord_type='lon')]
            else:
                yx = [int(event.ydata+0.5),
                      int(event.xdata+0.5)]
            self.plot_coherence_matrix4pixel(yx)



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Heresh Fattahi, 2013                             #
############################################################


import sys

import numpy as np

import mintpy.utils.readfile as readfile

################################################################################
USAGE = """
usage: correlation_with_dem.py data_file dem_file [y0:y1 x0:x1]

Calculates the correlation of the DEM and file

example:
  correlation_with_dem.py velocity_masked.h5 inputs/geometryRadar.h5
  correlation_with_dem.py velocity_masked.h5 radar_8rlks.hgt
"""


def usage():
    print(USAGE)
    return


################################################################################
def main(argv):
    try:
        File = argv[0]
    except:
        usage()
        sys.exit(1)

    try:
        demFile = argv[1]
    except:
        demFile = 'inputs/geometryRadar.h5'

    dem = readfile.read(demFile, datasetName='height')[0]
    data, atr = readfile.read(File)
    print('input file is '+atr['FILE_TYPE'])

    # Subset
    try:
        y0, y1 = (int(i) for i in argv[2].split(':'))
        x0, x1 = (int(i) for i in argv[3].split(':'))
        data = data[y0:y1, x0:x1]
        dem = dem[y0:y1, x0:x1]
    except:
        pass

    # Calculation
    dem = dem.flatten(1)
    data = data.flatten(1)
    ndx = ~np.isnan(data)
    C1 = np.zeros([2, len(dem[ndx])])
    C1[0][:] = dem[ndx]
    C1[1][:] = data[ndx]

    # Display
    print('-------------------------------------------')
    print('Correlation with the DEM:  %.2f' % np.corrcoef(C1)[0][1])
    print('-------------------------------------------')
    print('DEM info:')
    print('    Max height difference: %.2f m' % (np.max(dem[ndx])-np.min(dem[ndx])))
    print('    Average        height: %.2f m' % np.mean(dem[ndx]))
    print('    Height            Std: %.2f m' % np.std(dem[ndx]))
    return


################################################################################
if __name__ == '__main__':
    main(sys.argv[1:])



############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, 2018                               #
############################################################
# Recommended usage:
#   from mintpy.simulation import simulation as sim
#   OR from mintpy.simulation import defo_model as defo


import numpy as np
from matplotlib import pyplot as plt

from mintpy.utils import utils0 as ut0


def mogi(geometry, xloc, nu=0.25):
    """Computes surface displacements, strains, and tilts due to a Mogi source.
    Parameters: geometry : tuple of 4 float, Mogi source geometry: East, North, Depth, Volomn change in SI unit.
                xloc : 2D np.array in size of (2, num_pixel), surface coordinates in east/x and north/y
                nu : float, Poisson's ratio
    Returns:    u : 2D np.array of displacement in size of (3, num_pixel) for Ux, Uy, Uz
                e : 2D np.array of strains      in size of (3, num_pixel) for Exx, Exy, Eyy
                t : 2D np.array of tilts        in size of (2, num_pixel) for dUz/dx, dUz/dy
    Notes: depth denotes an unsigned length and should therefore always be given positive. Keep your units consistent!
    This is a python translation from mogi.m originally written by Peter Cervelli, May 1998.
    """
    xloc = np.array(xloc, np.float32).reshape(2, -1)

    # compute displacements
    num_pixel = xloc.shape[1]
    E = geometry[0] - xloc[0, :]
    N = geometry[1] - xloc[1, :]
    E2 = np.square(E)
    N2 = np.square(N)
    d2 = np.square(geometry[2])
    R = np.sqrt(d2 + E2 + N2)
    C = (nu - 1.) * geometry[3] / np.pi

    R3 = C * np.power(R, -3)
    displacement = np.zeros((3, num_pixel), np.float32)
    displacement[0, :] = np.multiply(E, R3)
    displacement[1, :] = np.multiply(N, R3)
    displacement[2, :] = np.multiply(-1 * geometry[2], R3)

    # compute strains (if necessary)
    R5 = C * np.power(R, -5)
    strain = np.zeros((3, num_pixel), np.float32)
    strain[0, :] = np.multiply(R5, (2. * E2 - N2 - d2))
    strain[1, :] = 3. * np.multiply(R5, np.multiply(E, N))
    strain[2, :] = np.multiply(R5, (2. * N2 - E2 - d2))

    # compute tilts
    tilt = np.zeros((2, num_pixel), np.float32)
    tilt[0, :] = -3. * np.multiply(R5, E) * geometry[2]
    tilt[1, :] = -3. * np.multiply(R5, N) * geometry[2]

    return displacement, strain, tilt


def mogi_los(shape, source_geom, resolution=60., scale=1., display=True):
    """Simulate 2D deformation caused by the overpress of a Mogi source underneath

    Parameters: shape: 2-tuple of int in (length, width) or 2D np.ndarray in size of (length, width) in np.bool_
                source_geom : 4-tuple of float, Mogi source geometry: East, North, Depth, Volomn change in SI unit.
    Returns:    dis_los: 2D np.ndarray in size of (length, width), deformation in LOS direction in meter
    """
    if isinstance(shape, np.ndarray):
        mask = np.multiply(np.array(shape != 0), ~np.isnan(shape))
        shape = mask.shape
    else:
        mask = np.ones(shape, np.bool_)

    length, width = shape
    yy, xx = np.mgrid[0:length:length*1j, 0:width:width*1j]
    yy *= resolution
    xx *= resolution
    xloc = np.vstack((xx.reshape(1, -1), yy.reshape(1, -1)))

    dis_map = mogi(source_geom, xloc)[0]
    dis_e = dis_map[0, :].reshape(length, width)
    dis_n = dis_map[1, :].reshape(length, width)
    dis_u = dis_map[2, :].reshape(length, width)

    dis_los = ut0.enu2los(dis_e, dis_n, dis_u,
                          inc_angle=34.,
                          head_angle=-168.)
    dis_los[mask == 0.] = np.nan
    dis_los *= scale

    if display:
        fig, ax = plt.subplots(1, 4, figsize=[10, 3], sharey=True)
        dmin = np.nanmin(dis_los)
        dmax = np.nanmax(dis_los)
        for i, fig_title in enumerate(['east','north','vertical']):
            ax[i].imshow(dis_map[i, :].reshape(length, width), vmin=dmin, vmax=dmax)
            ax[i].set_title(fig_title)
        im = ax[3].imshow(dis_los, vmin=dmin, vmax=dmax)
        ax[3].set_title('los - SenD')
        fig.subplots_adjust(right=0.90)
        cax = fig.add_axes([0.92, 0.25, 0.01, 0.5])
        cbar = fig.colorbar(im, cax=cax)
        cbar.set_label('Displacement [m]')
        plt.show()

    return dis_los



############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Zhang Yunjun, Xiaohua Xu, Mar 2021               #
############################################################


import glob
import os

import numpy as np

try:
    from osgeo import gdal
except ImportError:
    raise ImportError('Can not import gdal!')

from mintpy.utils import ptime, readfile, utils as ut, writefile


#########################################################################
def get_prm_files(ifg_dir):
    """Get the date1/2.prm files in the interferogram directory."""
    prm_files = sorted(glob.glob(os.path.join(ifg_dir, '*.PRM')))
    prm_files = [i for i in prm_files if os.path.splitext(os.path.basename(i))[0].isdigit()]
    if len(prm_files) != 2:
        raise FileExistsError('NO two *.prm files found!')
    return prm_files


def get_multilook_number(ifg_dir, fbases=['corr', 'phase', 'phasefilt', 'unwrap']):
    """Get the number of multilook in range and azimuth direction."""
    # grab an arbitrary file in radar-coordiantes
    rdr_files = [os.path.join(ifg_dir, f'{i}.grd') for i in fbases]
    if len(rdr_files) == 0:
        raise ValueError(f'No radar-coord files found in {ifg_dir} with suffix: {fbases}')

    # read step info into multilook number
    ds = gdal.Open(rdr_files[0], gdal.GA_ReadOnly)
    transform = ds.GetGeoTransform()
    rlooks = int(abs(transform[1]))
    alooks = int(abs(transform[5]))
    return alooks, rlooks


def get_lalo_ref(ifg_dir, prm_dict, fbases=['corr', 'phase', 'phasefilt', 'unwrap']):
    """Get the LAT/LON_REF1/2/3/4 from *_ll.grd file."""
    # grab an arbitrary file in geo-coordiantes
    geo_files = [os.path.join(ifg_dir, f'{i}_ll.grd') for i in fbases]
    if len(geo_files) == 0:
        print(f'WARNING: No geo-coord files found in {ifg_dir} with suffix: {fbases}')
        return prm_dict

    # read corners lat/lon info
    ds = gdal.Open(geo_files[0], gdal.GA_ReadOnly)
    transform = ds.GetGeoTransform()
    x_step = abs(transform[1])
    y_step = abs(transform[5]) * -1.
    if not (1e-7 < x_step < 1.):
        raise ValueError('File {} is NOT geocoded!')

    W = transform[0] - x_step / 2.
    N = transform[3] - y_step / 2.
    E = W + x_step * ds.RasterXSize
    S = N + y_step * ds.RasterYSize

    if prm_dict['ORBIT_DIRECTION'].upper().startswith('ASC'):
        prm_dict['LAT_REF1'] = str(S)
        prm_dict['LAT_REF2'] = str(S)
        prm_dict['LAT_REF3'] = str(N)
        prm_dict['LAT_REF4'] = str(N)
        prm_dict['LON_REF1'] = str(W)
        prm_dict['LON_REF2'] = str(E)
        prm_dict['LON_REF3'] = str(W)
        prm_dict['LON_REF4'] = str(E)
    else:
        prm_dict['LAT_REF1'] = str(N)
        prm_dict['LAT_REF2'] = str(N)
        prm_dict['LAT_REF3'] = str(S)
        prm_dict['LAT_REF4'] = str(S)
        prm_dict['LON_REF1'] = str(E)
        prm_dict['LON_REF2'] = str(W)
        prm_dict['LON_REF3'] = str(E)
        prm_dict['LON_REF4'] = str(W)

    return prm_dict


def get_slant_range_distance(ifg_dir, prm_dict, fbases=['corr', 'phase', 'phasefilt', 'unwrap']):
    """Get a constant slant range distance in the image center, for dataset in geo-coord."""
    # grab an arbitrary file in radar-coordiantes
    rdr_files = [os.path.join(ifg_dir, f'{i}.grd') for i in fbases]
    if len(rdr_files) == 0:
        raise ValueError(f'No radar-coord files found in {ifg_dir} with suffix: {fbases}')

    # read width from rdr_file
    ds = gdal.Open(rdr_files[0], gdal.GA_ReadOnly)
    width = ds.RasterXSize

    near_range = float(prm_dict['STARTING_RANGE'])
    range_pixel_size = float(prm_dict['RANGE_PIXEL_SIZE'])
    slant_range_dist = near_range + range_pixel_size * width / 2.
    prm_dict['SLANT_RANGE_DISTANCE'] = slant_range_dist

    return prm_dict


#########################################################################
def extract_gmtsar_metadata(unw_file, template_file, rsc_file=None, update_mode=True):
    """Extract metadata from GMTSAR interferogram stack."""

    # update_mode: check existing rsc_file
    if update_mode and ut.run_or_skip(rsc_file, in_file=unw_file, readable=False) == 'skip':
        return readfile.read_roipac_rsc(rsc_file)

    ifg_dir = os.path.dirname(unw_file)

    # 1. read *.PRM file
    prm_file = get_prm_files(ifg_dir)[0]
    meta = readfile.read_gmtsar_prm(prm_file)
    meta['PROCESSOR'] = 'gmtsar'

    # 2. read template file: HEADING, ORBIT_DIRECTION
    template = readfile.read_template(template_file)
    for key in ['HEADING', 'ORBIT_DIRECTION']:
        if key in template.keys():
            meta[key] = template[key].lower()
        else:
            raise ValueError('Attribute {} is missing! Please manually specify it in the template file.')

    # 3. grab A/RLOOKS from radar-coord data file
    meta['ALOOKS'], meta['RLOOKS'] = get_multilook_number(ifg_dir)
    meta['AZIMUTH_PIXEL_SIZE'] *= meta['ALOOKS']
    meta['RANGE_PIXEL_SIZE'] *= meta['RLOOKS']

    # 4. grab LAT/LON_REF1/2/3/4 from geo-coord data file
    meta = get_lalo_ref(ifg_dir, meta)

    # 5. grab X/Y_FIRST/STEP from unw_file if in geo-coord
    ds = gdal.Open(unw_file, gdal.GA_ReadOnly)
    transform = ds.GetGeoTransform()
    x_step = abs(transform[1])
    y_step = abs(transform[5]) * -1.
    if 1e-7 < x_step < 1.:
        meta['X_STEP'] = x_step
        meta['Y_STEP'] = y_step
        meta['X_FIRST'] = transform[0] - x_step / 2.
        meta['Y_FIRST'] = transform[3] - y_step / 2.
        # constrain longitude within (-180, 180]
        if meta['X_FIRST'] > 180.:
            meta['X_FIRST'] -= 360.

    # 6. extra metadata for the missing geometry dataset: SLANT_RANGE_DISTANCE / INCIDENCE_ANGLE
    # for dataset in geo-coordinates
    if 'Y_FIRST' in meta.keys():
        meta = get_slant_range_distance(ifg_dir, meta)
        Re = float(meta['EARTH_RADIUS'])
        H = float(meta['HEIGHT'])
        Rg = float(meta['SLANT_RANGE_DISTANCE'])
        Inc = (np.pi - np.arccos((Re**2 + Rg**2 - (Re+H)**2) / (2*Re*Rg))) * 180./np.pi
        meta['INCIDENCE_ANGLE'] = Inc

    # convert all value to string format
    for key, value in meta.items():
        meta[key] = str(value)

    # write to .rsc file
    meta = readfile.standardize_metadata(meta)
    if rsc_file:
        print('writing ', rsc_file)
        os.makedirs(os.path.dirname(rsc_file), exist_ok=True)
        writefile.write_roipac_rsc(meta, rsc_file)

    return meta


def prepare_geometry(geom_files, meta, update_mode=True):
    """Prepare .rsc file for all geometry files."""
    num_file = len(geom_files)
    if num_file == 0:
        raise FileNotFoundError('NO geometry file found!')

    # write .rsc file for each geometry file
    for geom_file in geom_files:
        # copy over the common metadata
        geom_meta = {}
        for key, value in meta.items():
            geom_meta[key] = value

        # update from .grd file
        geom_meta.update(readfile.read_gdal_vrt(geom_file))

        # write .rsc file
        rsc_file = geom_file+'.rsc'
        writefile.write_roipac_rsc(
            geom_meta,
            rsc_file,
            update_mode=update_mode,
            print_msg=True,
        )

    return


def prepare_stack(unw_files, meta, update_mode=True):
    """Prepare .rsc file for all unwrapped interferogram files."""
    num_file = len(unw_files)
    if num_file == 0:
        raise FileNotFoundError('NO unwrapped interferogram file found!')

    # write .rsc file for each interferogram file
    prog_bar = ptime.progressBar(maxValue=num_file)
    for i, unw_file in enumerate(unw_files):
        ifg_dir = os.path.dirname(unw_file)
        ifg_meta = {}

        # copy over the common metadata
        for key, value in meta.items():
            ifg_meta[key] = value

        # update from .grd file
        ifg_meta.update(readfile.read_gdal_vrt(unw_file))

        # add DATE12
        prm_files = get_prm_files(ifg_dir)
        date1, date2 = (os.path.splitext(os.path.basename(i))[0] for i in prm_files)
        ifg_meta['DATE12'] = f'{ptime.yymmdd(date1)}-{ptime.yymmdd(date2)}'

        # and P_BASELINE_TOP/BOTTOM_HDR
        baseline_file = os.path.join(ifg_dir, 'baseline.txt')
        if os.path.isfile(baseline_file):
            bDict = readfile.read_template(baseline_file)
            ifg_meta['P_BASELINE_TOP_HDR'] = bDict['B_perpendicular']
            ifg_meta['P_BASELINE_BOTTOM_HDR'] = bDict['B_perpendicular']
        else:
            ifg_meta['P_BASELINE_TOP_HDR'] = '0'
            ifg_meta['P_BASELINE_BOTTOM_HDR'] = '0'
            msg = f'WARNING: No baseline file found in: {baseline_file}. '
            msg += 'Set P_BASELINE* to 0 and continue.'
            print(msg)

        # write .rsc file
        rsc_file = unw_file+'.rsc'
        writefile.write_roipac_rsc(
            ifg_meta,
            rsc_file,
            update_mode=update_mode,
            print_msg=False,
        )

        prog_bar.update(i+1, suffix=f'{date1}_{date2}')
    prog_bar.close()


#########################################################################
def prep_gmtsar(inps):
    # read file path from template file
    template = readfile.read_template(inps.template_file)
    inps.unw_files = sorted(glob.glob(template['mintpy.load.unwFile']))
    inps.cor_files = sorted(glob.glob(template['mintpy.load.corFile']))
    inps.dem_file = glob.glob(template['mintpy.load.demFile'])[0]

    # extract common metadata
    rsc_file = os.path.join(inps.mintpy_dir, 'inputs/data.rsc')
    meta = extract_gmtsar_metadata(
        unw_file=inps.unw_files[0],
        template_file=inps.template_file,
        rsc_file=rsc_file,
        update_mode=inps.update_mode,
    )

    # prepare metadata for geometry files
    prepare_geometry(
        geom_files=[inps.dem_file],
        meta=meta,
        update_mode=inps.update_mode,
    )

    # prepare metadata for interferogram files
    prepare_stack(
        unw_files=inps.unw_files,
        meta=meta,
        update_mode=inps.update_mode,
    )

    print('Done.')
    return



#!/usr/bin/env python3
############################################################
# Program is part of MintPy                                #
# Copyright (c) 2013, Zhang Yunjun, Heresh Fattahi         #
# Author: Antonio Valentino, Zhang Yunjun, Aug 2022        #
############################################################


import os
import sys

from mintpy.utils.arg_utils import create_argument_parser

# from mintpy.objects.sensor import SENSOR_NAMES
# copy as below to avoid importing the non-empty mintpy.objects.__init__.py
SENSOR_NAMES = [
    'alos',
    'alos2',
    'csk',
    'env',
    'ers',
    'gf3',
    'jers',
    'ksat5',
    'ni',
    'rs1',
    'rs2',
    'rcm',
    'sen',
    'tsx',
    'uav',
]


##################################################################################################
NOTE = """
  For each interferogram, including unwrapped/wrapped interferograms and coherence, 3 metadata files are required:
  1) reference .par file, e.g. 130118_4rlks.amp.par
  2) secondary .par file, e.g. 130129_4rlks.amp.par
  3) interferogram .off file, e.g. 130118-130129_4rlks.off

  Other metadata files are recommended and can be generated from the above 3 if not existed, more specifically:
  4) baseline files, e.g. 130118-130129_4rlks.baseline and 130118-130129_4rlks.base_perp,
      which can be generated from file 1-3 with Gamma command base_orbit and base_perp.
  5) corner files, e.g. 130118_4rlks.amp.corner_full and 130118_4rlks.amp.corner,
      which can be generated from file 1 with Gamma command SLC_corners.

  This script will read all these files (generate 4 and 5 if not existed), merge them into one, convert their name from
  Gamma style to ROI_PAC style, and write to an metadata file, same name as input binary data file with suffix .rsc,
  e.g. diff_filt_HDR_130118-130129_4rlks.unw.rsc


  For DEM file in radar/geo coordinates (.hgt_sim or .rdc.dem / .utm.dem) and
      lookup table file for geocoding (.UTM_TO_RDC), 2 metadata files are required:
  1) .par      file, for DEM in geo   coordinates and lookup table, e.g.: sim_150911_4rlks.utm.dem.par
  2) .diff_par file, for DEM in radar coordinates, e.g. sim_150911_4rlks.diff_par


  Here is an example of how your Gamma files should look like:
  Before loading:
      For each interferogram, 5 files are needed:
          130118-130129_4rlks.off
          130118_4rlks.amp.par
          130129_4rlks.amp.par
          filt_130118-130129_4rlks.cor
          diff_130118-130129_4rlks.unw
      For each dataset, only one sim* folder with 5 files are needed,
          sim_150911_4rlks.UTM_TO_RDC
          sim_150911_4rlks.diff_par
          sim_150911_4rlks.hgt_sim or sim_150911.rdc.dem
          sim_150911_4rlks.utm.dem
          sim_150911_4rlks.utm.dem.par
  After running prep_gamma.py:
      For each interferogram:
          130118-130129_4rlks.base_perp
          130118-130129_4rlks.baseline
          130118-130129_4rlks.off
          130118_4rlks.ramp.corner
          130118_4rlks.ramp.corner_full
          130118_4rlks.ramp.par
          130129_4rlks.ramp.par
          filt_130118-130129_4rlks.cor
          filt_130118-130129_4rlks.cor.rsc
          diff_130118-130129_4rlks.unw
          diff_130118-130129_4rlks.unw.rsc
      For the geometry files in each dataset:
          sim_150911_4rlks.UTM_TO_RDC
          sim_150911_4rlks.UTM_TO_RDC.rsc
          sim_150911_4rlks.diff_par
          sim_150911_4rlks.rdc.dem      or sim_150911_4rlks.hgt_sim
          sim_150911_4rlks.rdc.dem.rsc  or sim_150911_4rlks.hgt_sim.rsc
          sim_150911_4rlks.utm.dem
          sim_150911_4rlks.utm.dem.par
          sim_150911_4rlks.utm.dem.rsc

  Notes: both - and _ are supported;
         both YYMMDD and YYYYMMDD naming are also supported;
         if no multilooking applied, do not add "_4rlks" in your file names.
"""

EXAMPLE = """example:
  prep_gamma.py  diff_filt_HDR_20130118_20130129_4rlks.unw
  prep_gamma.py  interferograms/*/diff_*rlks.unw --sensor sen
  prep_gamma.py  interferograms/*/filt_*rlks.cor
  prep_gamma.py  interferograms/*/diff_*rlks.int
  prep_gamma.py  sim_20150911_20150922.hgt_sim
  prep_gamma.py  sim_20150911_20150922.utm.dem
  prep_gamma.py  sim_20150911_20150922.UTM_TO_RDC
"""


def create_parser(subparsers=None):
    synopsis = 'Prepare attributes file for Gamma product.'
    epilog = EXAMPLE
    name = __name__.split('.')[-1]
    parser = create_argument_parser(
        name, synopsis=synopsis, description=synopsis+NOTE, epilog=epilog, subparsers=subparsers)

    parser.add_argument('file', nargs='+', help='Gamma file(s)')
    parser.add_argument('--sensor', dest='sensor', type=str, choices=SENSOR_NAMES,
                        help='SAR sensor')
    return parser


def cmd_line_parse(iargs=None):
    # parse
    parser = create_parser()
    inps = parser.parse_args(args=iargs)

    # import
    from mintpy.utils import utils1 as ut

    # check
    inps.file = ut.get_file_list(inps.file, abspath=True)
    inps.file_ext = os.path.splitext(inps.file[0])[1]

    # check: input file extension
    ext_list = ['.unw', '.cor', '.int', '.dem', '.hgt_sim', '.UTM_TO_RDC']
    if inps.file_ext not in ext_list:
        msg = f'unsupported input file extension: {inps.file_ext}'
        msg += f'\nsupported file extensions: {ext_list}'
        raise ValueError(msg)

    return inps


##################################################################################################
def main(iargs=None):
    # parse
    inps = cmd_line_parse(iargs)

    # import
    from mintpy.prep_gamma import prep_gamma

    # run
    prep_gamma(inps)


###################################################################################################
if __name__ == '__main__':
    main(sys.argv[1:])
